import numpy as np
#import matplotlib
#matplotlib.use('Agg')
import pylab as pl
from astropy.cosmology import Planck15
from scipy import interpolate
import mmf
import scipy.signal as sp
import pymaster as nmt
import time
import os
import expt
from numba import jit
import scipy.signal as sg
import healpy as hp
import inpaint
import scipy
import scipy.stats as st

class pixel:

    def __init__(self,nx,dx,ny=None,dy=None):

        if ny == None:

            ny = nx

        if dy == None:

            dy = dx

        self.nx = nx
        self.dx = dx
        self.ny = ny
        self.dy = dy

    def get_dx_phys(self,cosmology,z):

        return cosmology.angular_diameter_distance(z).value*self.dx  #in Mpc

    def get_dy_phys(self,cosmology,z):

        return cosmology.angular_diameter_distance(z).value*self.dy  #in Mpc

class rmap:

    def __init__(self,pix,map_value=None,):

        self.nx = pix.nx
        self.dx = pix.dx
        self.ny = pix.ny
        self.dy = pix.dy
        self.pix = pix

        if map_value is None:

            map_value = np.zeros((self.nx,self.ny))

        self.map_value = map_value

    def get_x_coord_map_wrt_centre(self,theta_x=0.):#actually y coordinate, theta_x in radians

        x_coord = (np.arange(0,self.nx)-self.nx*0.5+0.5)*self.dx + theta_x
        x_coord_map = np.transpose(np.tile(x_coord,(self.ny,1)))

        return x_coord_map

    def get_y_coord_map_wrt_centre(self,theta_y=0.):

        y_coord = (np.arange(0,self.ny)-self.ny*0.5+0.5)*self.dy + theta_y
        y_coord_map = np.tile(y_coord,(self.nx,1))

        return y_coord_map

    def get_x_coord_map_wrt_origin(self):#actually y coordinate

        x_coord = (np.arange(0,self.nx)+0.5)*self.dx
        x_coord_map = np.transpose(np.tile(x_coord,(self.ny,1)))

        return x_coord_map

    def get_y_coord_map_wrt_origin(self):

        y_coord = (np.arange(0,self.ny)+0.5)*self.dy
        y_coord_map = np.tile(y_coord,(self.nx,1))

        return y_coord_map

    def get_distance_map_wrt_centre(self,theta_misc=[0.,0.]):

        (theta_x,theta_y) = theta_misc

        return np.sqrt(self.get_x_coord_map_wrt_centre(theta_x)**2+self.get_y_coord_map_wrt_centre(theta_y)**2)

    def get_fft(self):

        return get_fft(self.map_value,self.pix)

    def get_ifft(self):

        return get_ifft(self.map_value,self.pix)

    def get_lxly(self):

        return np.meshgrid(np.fft.fftfreq(self.ny,self.dy)*2.*np.pi,np.fft.fftfreq(self.nx,self.dx)*2.*np.pi)

    def get_ell(self):

        lx, ly = self.get_lxly()

        return np.sqrt(lx**2 + ly**2)

    def convolve_gaussian(self,fwhm_arcmin):

        return get_gaussian_convolution(self.map_value,fwhm_arcmin,self.pix)

    def filter_lrange(self,lrange):

        print(self.get_ell().shape)

        ell =  self.get_ell().reshape(self.pix.nx*self.pix.ny)
        map_filtered = self.map_value.reshape(self.pix.nx*self.pix.ny)

        print(map_filtered.shape)
        #map_filtered[np.where((ell > lrange[1]) & (ell < lrange[0]))] = 0.
        map_filtered.reshape((self.pix.nx,self.pix.ny))

        return map_filtered



class tmap:

    def __init__(self,pix,n_freq,map_value=None):

        self.nx = pix.nx
        self.dx = pix.dx
        self.ny = pix.ny
        self.dy = pix.dy
        self.n_freq = n_freq
        self.pix = pix

        if map_value is None:

            map_value = np.zeros((self.nx,self.ny,self.n_freq))

        self.map_value = map_value

    def get_x_coord_map_wrt_centre(self,theta_x=0.):

        x_coord = (np.arange(0,self.nx)-self.nx*0.5+0.5)*self.dx + theta_x
        x_coord_map = np.tile(x_coord,(self.nx,1))

        return x_coord_map

    def get_y_coord_map_wrt_centre(self,theta_y=0.):

        y_coord = (np.arange(0,self.ny)-self.ny*0.5+0.5)*self.dy + theta_y
        y_coord_map = np.tile(y_coord,(self.ny,1))

        return np.transpose(y_coord_map)

    def get_distance_map_wrt_centre(self,theta_misc=[0.,0.]):

        (theta_x,theta_y) = theta_misc

        return np.sqrt(self.get_x_coord_map_wrt_centre(theta_x)**2+self.get_y_coord_map_wrt_centre(theta_y)**2)

    def get_fft(self):

        fmap_fft = np.zeros((self.n_freq,self.nx,self.ny))

        for i in range(0,self.n_freq):

            fmap_fft[:,:,i] = get_fft(self.map_value[:,:,i],self.pix)

        return fmap_fft

    def get_ifft(self):

        fmap_ifft = np.zeros((self.n_freq,self.nx,self.ny))

        for i in range(0,self.n_freq):

            fmap_ifft[:,:,i] = get_ifft(self.map_value[:,:,i],self.pix)

        return fmap_ifft

    def get_lxly(self):

        return np.meshgrid(np.fft.fftfreq(self.ny,self.dy)*2.*np.pi,np.fft.fftfreq(self.nx,self.dx)*2.*np.pi)

    def get_ell(self):

        lx, ly = self.get_lxly()

        return np.sqrt(lx**2 + ly**2)

    def convolve_gaussian(self,fwhm_arcmin_vec):

        fmap_convolved = np.zeros((self.nx,self.ny,self.n_freq))

        for i in range(0,self.n_freq):

            fmap_convolved[:,:,i] = get_gaussian_convolution(self.map_value[:,:,i],fwhm_arcmin_vec[i],self.pix)

        return fmap_convolved

    def filter_lrange(self,lrange):

        tmap_filtered = np.zeros(self.map_value.shape)

        for i in range(0,self.map_value.shape[2]):

            tmap_filtered[:,:,i] = rmap(self.pix,map_value=self.map_value[:,:,i]).filter_lrange(lrange)

        return tmap_filtered


def get_fft(map_value,pix):

    return np.fft.fft2(map_value)*np.sqrt((pix.dx*pix.dy)/(pix.nx*pix.ny))#same convention as quicklens

def get_ifft(map_value,pix):

    return np.fft.ifft2(map_value)*np.sqrt((pix.nx*pix.ny)/(pix.dx*pix.dy)) #same convention as quicklens

def fftconvolve(map1,map2,pix):

    ret = get_ifft(get_fft(map1,pix)*get_fft(map2,pix),pix).real
    ret2 = np.zeros((pix.nx,pix.ny))

    for i in range(0,pix.nx):

        for j in range(0,pix.ny):

            ret2[i,j] = ret[i-pix.nx//2,j-pix.ny//2]

    return ret2


def get_fft_f(tmap,pix):

    n_freq = tmap.shape[2]
    ret = np.zeros(tmap.shape,dtype=complex)

    for i in range(0,n_freq):

        ret[:,:,i] = get_fft(tmap[:,:,i],pix)

    return ret

def get_ifft_f(tmap,pix):

    n_freq = tmap.shape[2]
    ret = np.zeros(tmap.shape,dtype=complex)

    for i in range(0,n_freq):

        ret[:,:,i] = get_ifft(tmap[:,:,i],pix)

    return ret

def filter_fft(map_fft,pix,ell_filter):

    [lmin,lmax] = ell_filter
    indices = np.where((rmap(pix).get_ell() < lmin) |  (rmap(pix).get_ell() > lmax))
    map_fft[indices] = 0.

    return map_fft

def filter_fft_f(map_fft,pix,ell_filter):

    for i in range(0,map_fft.shape[2]):

        map_fft[:,:,i] = filter_fft(map_fft[:,:,i],pix,ell_filter)

    return map_fft

def filter_tmap(tmap,pix,ell_filter):

    return get_ifft_f(filter_fft_f(get_fft_f(tmap,pix),pix,ell_filter),pix).real

def filter_map(map,pix,ell_filter):

    return get_ifft(filter_fft(get_fft(map,pix),pix,ell_filter),pix).real

def filter_cov(cov,pix,ell_filter):

    [lmin,lmax] = ell_filter
    indices0,indices1 = np.where((rmap(pix).get_ell() < lmin) |  (rmap(pix).get_ell() > lmax))
    cov[indices0,indices1,:,:] = 0.

    return cov

def get_gaussian_convolution(map_value,fwhm_arcmin,pix):

    map_fft = get_fft(map_value,pix)
    ell = rmap(pix).get_ell()
    beam_fft = get_bl(fwhm_arcmin,ell)
    map_fft_convolved = get_ifft(map_fft*beam_fft,pix).real

    return map_fft_convolved

def get_gaussian_convolution_fft(map_fft,fwhm_arcmin,pix):

    ell = rmap(pix).get_ell()
    beam_fft = get_bl(fwhm_arcmin,ell)
    map_fft_convolved = map_fft*beam_fft

    return map_fft_convolved

def get_gaussian_deconvolution(map_value,fwhm_arcmin,pix):

    map_fft = get_fft(map_value,pix)
    ell = rmap(pix).get_ell()
    beam_fft = get_bl(fwhm_arcmin,ell)
    map_fft_convolved = get_ifft(map_fft/beam_fft,pix).real

    return map_fft_convolved

def get_convolution_isotropic(map_value,kernel,pix):

    ell_kernel,fft_kernel = kernel
    map_fft = get_fft(map_value,pix)
    ell = rmap(pix).get_ell()
    beam_fft = np.interp(ell,ell_kernel,fft_kernel)
    map_fft_convolved = get_ifft(map_fft*beam_fft,pix).real

    return map_fft_convolved

def get_convolution_isotropic_fft(map_fft,kernel,pix):

    ell_kernel,fft_kernel = kernel
    ell = rmap(pix).get_ell()
    beam_fft = np.interp(ell,ell_kernel,fft_kernel)
    map_fft_convolved = map_fft*beam_fft

    return map_fft_convolved


def get_deconvolution_isotropic(map_value,kernel,pix):

    ell_kernel,fft_kernel = kernel
    map_fft = get_fft(map_value,pix)
    ell = rmap(pix).get_ell()
    beam_fft = np.interp(ell,ell_kernel,fft_kernel)
    map_fft_convolved = get_ifft(map_fft/beam_fft,pix).real

    return map_fft_convolved

def convolve(map1,map2,pix):

    return get_ifft(get_fft(map1,pix)*get_fft(map2,pix),pix).real

def get_bl(fwhm_arcmin,ell):

    return np.exp(-(fwhm_arcmin*np.pi/180./60.)**2/(16.*np.log(2.))*ell*(ell+1.))

#theta in rad

def get_beam_real(fwhm_arcmin,theta):

    sigma_arcmin = fwhm_arcmin/(2.*np.sqrt(2.*np.log(2)))

    return eval_gaussian(theta,sigma_arcmin/60./180.*np.pi)

def eval_gaussian(x,sigma):

    return st.norm.pdf(x,scale=sigma)

def eval_beam_real_map(pix,fwhm_arcmin):

    theta_map = rmap(pix).get_distance_map_wrt_centre()

    return get_beam_real(fwhm_arcmin,theta_map)

def eval_real_beam_real_map(pix,beam_data):

    ell_map = rmap(pix).get_ell()

    ell_beam,beam_fft = beam_data

    fwhm = expt.planck_specs().FWHM[0]
    beam_gaussian = get_bl(fwhm,ell_map)
    beam_gaussian[0,0] = 0.

    beam_real_space = eval_beam_real_map(pix,fwhm)
    beam_fft_gaussian = get_fft(beam_real_space,pix)

    """
    pl.imshow(beam_gaussian.imag)
    pl.show()
    pl.imshow(beam_fft.imag)
    pl.show()
    pl.imshow(beam_gaussian.real/beam_fft.real)
    pl.show()
    """

    beam_fft = np.interp(ell_map,ell_beam,beam_fft)*np.sign(beam_fft_gaussian)

    return get_ifft(beam_fft,pix)

def get_bl_vec(fwhm_arcmin,lmax):
    """ returns the map-level transfer function for a symmetric Gaussian beam.
         * fwhm_arcmin      = beam full-width-at-half-maximum (fwhm) in arcmin.
         * lmax             = maximum multipole.
    """
    ls = np.arange(0,lmax+1)

    return (ls,get_bl(fwhm_arcmin,ls))

#Beam-deconvolved noise power spectrum

def get_nl(n_lev,fwhm_arcmin,ell):

    return (n_lev*np.pi/180./60.)**2/get_bl(fwhm_arcmin,ell)**2

def get_nl_vec(n_lev,fwhm_arcmin,lmax):

    ls = np.arange(0,lmax+1)

    return (ls,get_nl(n_lev,fwhm_arcmin,ls))


def rfft2_to_fft2(pix,rfft):

    ny = pix.ny
    fft = np.zeros((pix.nx,pix.ny),dtype=complex)
    fft[:,0:(ny//2+1)] = rfft[:,:]
    fft[0,(ny//2+1):]  = np.conj(rfft[0,1:ny//2][::-1])
    fft[1:,(ny//2+1):]  = np.conj(rfft[1:,1:ny//2][::-1,::-1])

    return fft


def nl(noise_uK_arcmin, fwhm_arcmin, lmax):
    """ returns the beam-deconvolved noise power spectrum in units of uK^2 for
         * noise_uK_arcmin = map noise level in uK.arcmin
         * fwhm_arcmin     = beam full-width-at-half-maximum (fwhm) in arcmin.
         * lmax            = maximum multipole.
    """
    return (noise_uK_arcmin*np.pi/180./60.)**2/bl(fwhm_arcmin,lmax)**2

def get_noise_map(pix,n_lev): #white noise

    sigma_pixel = n_lev/(180.*60./np.pi*pix.dx)

    return get_noise_map_sigma_pix(pix,sigma_pixel)


def get_noise_tmap(pix,n_lev_vec): #generates white noise frequency maps

    tmap = np.zeros((pix.nx,pix.ny,len(n_lev_vec)))

    for i in range(0,len(n_lev_vec)):

        tmap[:,:,i] = get_noise_map(pix,n_lev_vec[i])

    return tmap

def get_noise_tmap_Planck(pix,ptf_flag=False):

    noise = get_noise_tmap(pix,expt.planck_specs().noise_levels)

    if ptf_flag == True:

        exp = expt.planck_specs()

        for i in range(0,exp.n_freqs):

            noise[:,:,i] = get_convolution_isotropic(noise[:,:,i],exp.get_ptf(),pix)

    return noise

def get_noise_tmap_exp(pix,exp,ptf_flag=False):

    noise = get_noise_tmap(pix,exp.noise_levels)

    if ptf_flag == True:

        for i in range(0,exp.n_freqs):

            noise[:,:,i] = get_convolution_isotropic(noise[:,:,i],exp.get_ptf(),pix)

    return noise

def get_deconvolved_noise_map(pix,n_lev,fwhm_arcmin,lmax):

    ell,spec = get_nl_vec(n_lev,fwhm_arcmin,lmax)

    return get_realisation_from_spec_real(pix,ell,spec)

def get_noise_map_sigma_pix(pix,sigma_pixel):

    return np.random.standard_normal((pix.nx,pix.ny))*sigma_pixel

def get_nlev_from_sigma_pixel(pix,sigma_pixel):

    return sigma_pixel*(180.*60./np.pi*np.sqrt(pix.dx*pix.dy))


def get_cl_noise(ell,my_sz_data):

    return np.interp(ell,my_sz_data.l_eff,my_sz_data.cl_all)

def cl_to_map(ell,cl,pix,lmax=None):

    ell_map = rmap(pix).get_ell()
    ret = np.interp(ell_map,ell,cl)

    if lmax != None:

        indices = np.where(ell_map > lmax)
        ret[indices] = 0.
        print(np.where(ret==0.))

    return ret

def cl_binned_to_map(ell_bin,cl_bin,pix):

    ell = np.arange(ell_bin[0],ell_bin[-1])
    cl = np.interp(ell,ell_bin,cl_bin)
    ell_map = rmap(pix).get_ell()

    return np.interp(ell_map,ell,cl)

def get_noise_cl_map(pix):

    data = sz_data()
    ell_map = rmap(pix).get_ell()

    return get_cl_noise(ell_map,data)

def get_noise_cl_map_gaussian(pix,n_lev,fwhm_arcmin):

    ell_map = rmap(pix).get_ell()
    noise = get_nl(n_lev,fwhm_arcmin,ell_map)

    return noise

def get_cross_spectrum(pix,map1,map2=None,ell=None): #deprecated

    if ell == None:

        ell = rmap(pix).get_ell()

    if map2.all() == None:

        map2 = map1

    map1 = get_fft(map1,pix).flatten()
    map2 = get_fft(map2,pix).flatten()
    ell = ell.flatten()
    idx_sort = np.argsort(ell)
    map1_sorted = map1[idx_sort]
    map2_sorted = map2[idx_sort]
    ell_sorted = ell[idx_sort]
    ell_new, idx_start, count = np.unique(ell_sorted, return_counts=True, return_index=True)
    cross_spec = np.zeros(len(ell_new))

    for i in range(0,len(ell_new)):

        mapps1 = map1_sorted[idx_start[i]:idx_start[i]+count[i]]
        mapps2 = map2_sorted[idx_start[i]:idx_start[i]+count[i]]
        cross_spec[i] = np.mean(mapps1*np.conjugate(mapps2)).real

    return (ell_new,cross_spec)



class ps_mask:

    def __init__(self,pix,n_source,r_source_arcmin):

        self.pix = pix
        self.n_source = n_source
        self.r_source_arcmin = r_source_arcmin

    def get_source_coords(self): #with respect to the centre of the map

        source_coords = np.random.rand(self.n_source,2)-0.5
        source_coords[:,0] = source_coords[:,0]*self.pix.nx*self.pix.dx
        source_coords[:,1] = source_coords[:,1]*self.pix.ny*self.pix.dy

        return source_coords

    def get_mask_map(self,source_coords=None):

        if source_coords is None:

            source_coords = self.get_source_coords()

        mask = np.ones((self.pix.nx,self.pix.ny))

        for i in range(0,self.n_source):

            distances = rmap(self.pix).get_distance_map_wrt_centre(theta_misc=source_coords[i,:])
            r_source = self.r_source_arcmin/60./180.*np.pi
            indices = np.where(distances <= r_source)
            mask[indices] = 0.

        return mask

    def get_mask_map_t(self,n_freq,source_coords=None):

        mask = np.zeros((self.pix.nx,self.pix.ny,n_freq))
        mask_map = self.get_mask_map(source_coords=source_coords)

        for i in range(n_freq):

            mask[:,:,i] = mask_map

        return mask

def clone_map_freq(map1,n_freq):

    ret = np.zeros((map1.shape[0],map1.shape[1],n_freq))

    for i in range(0,n_freq):

        ret[:,:,i] = map1

    return ret

def get_mask_apod(pix,alpha=0.1):

    line1 = sp.tukey(pix.nx,alpha)
    line2 = sp.tukey(pix.ny,alpha)

    return np.outer(line1,line2)

def get_mask_apod_freq(pix,n_freq=6,alpha=0.1):

    mask = np.zeros((pix.nx,pix.ny,n_freq))

    for i in range(0,n_freq):

        mask[:,:,i] = get_mask_apod(pix,alpha=alpha)

    return mask

def get_mask_square(nx,buffer,ny=None):

    if ny is None:

        ny = nx

    mask = np.zeros((nx,ny))
    mask[buffer:nx-buffer,buffer:ny-buffer] = 1.

    return mask

def get_theta_cart(theta_misc,pix):

    theta_x_misc,theta_y_misc = theta_misc
    theta_x = pix.nx/2*pix.dx - theta_y_misc
    theta_y = pix.nx/2*pix.dx + theta_x_misc
    theta_cart = [theta_x,theta_y]

    return theta_cart

def get_theta_misc(theta_cart,pix):

    [theta_x,theta_y] = theta_cart
    theta_y_misc = pix.nx/2*pix.dx - theta_x
    theta_x_misc = -pix.nx/2*pix.dx + theta_y
    theta_misc = [theta_x_misc,theta_y_misc]

    return theta_misc

def get_ij_from_theta(theta_x,theta_y,pix):

    j = np.floor(theta_x/pix.dx).astype("int")
    i = (pix.ny-np.ceil(theta_y/pix.dy)).astype("int")

    return i,j

def get_theta_from_ij(i,j,pix):

    theta_x = j*pix.dx
    theta_y = (pix.ny-i)*pix.dy

    return theta_x,theta_y


@jit
def get_buffered_mask(pix,mask_input,buffer_arcmin):

    mask_output = np.ones(mask_input.shape)
    buffer_pix = int(round(buffer_arcmin/60./180.*np.pi/pix.dx))+1

    for i in range(0,mask_input.shape[0]):

        for j in range(0,mask_input.shape[1]):


            theta_x = j*pix.dy
            theta_y = (pix.nx-i)*pix.dx
            theta_cart = [theta_x,theta_y]
            distances = rmap(pix).get_distance_map_wrt_centre(theta_misc=get_theta_misc(theta_cart,pix))

            imin = i-buffer_pix
            imax = i+buffer_pix
            jmin = j-buffer_pix
            jmax = j+buffer_pix

            if imin < 0:

                imin = 0

            if imax >= pix.nx:

                imax = pix.nx-1

            if jmin < 0:

                jmin = 0

            if jmax >= pix.nx:

                jmax = pix.nx-1

            distances_select = distances[imin:imax,jmin:jmax]
            mask_select = mask_input[imin:imax,jmin:jmax]

            indices = np.where(distances_select <= buffer_arcmin/60./180.*np.pi)
            mask_select = mask_select[indices]

            if 0. in mask_select:

                mask_output[i,j] = 0.

    return mask_output


def get_buffered_mask_fft(pix,mask_input,buffer_arcmin):

    buffer_rad = buffer_arcmin/60./180.*np.pi
    buffer_pix = int(round(buffer_arcmin/60./180.*np.pi/pix.dx))+1

    nx_padded = pix.nx+4*buffer_pix
    ny_padded = pix.ny+4*buffer_pix

    mask_input_padded = np.ones((nx_padded,ny_padded))
    mask_input_padded[2*buffer_pix:nx_padded-2*buffer_pix,2*buffer_pix:ny_padded-2*buffer_pix] = mask_input

    pix_padded = pixel(nx_padded,pix.dx,ny=ny_padded,dy=pix.dy)

    kernel = np.zeros((pix_padded.nx,pix_padded.ny))
    distances = rmap(pix_padded).get_distance_map_wrt_centre()
    indices = np.where(distances <= buffer_rad)
    kernel[indices] = 1.
    kernel_area = np.sum(kernel)

    mask_convolved = np.around(sg.fftconvolve(mask_input_padded,kernel,mode='same')/kernel_area,decimals=4)[2*buffer_pix:nx_padded-2*buffer_pix,2*buffer_pix:ny_padded-2*buffer_pix]

    mask_output = np.ones(mask_input.shape)
    mask_output[np.where(mask_convolved != 1.)] = 0.

    return mask_output

#apotype: C1, C2 and Smooth

def get_apodised_mask(pix,mask_input,apotype="C1",aposcale=0.2):

    mask_input[0,:] = 0.
    mask_input[:,0] = 0.
    mask_input[-1,:] = 0.
    mask_input[:,-1] = 0.

    return nmt.mask_apodization_flat(mask_input,pix.nx*pix.dx,pix.ny*pix.dy,aposcale,apotype=apotype)

def get_fsky_criterion_mask(pix,mask_select,nside_tile,criterion=0.3):

    area_tile = hp.pixelfunc.nside2pixarea(nside_tile)
    area_select = np.sum(mask_select)*pix.dx*pix.dy
    frac = area_select/area_tile
    print("Tile frac",frac)

    if frac < criterion:

        mask_select = np.zeros((pix.nx,pix.ny))

    return mask_select

def mask_sigma(map,sigma,inpaint_flag=False,n_inpaint=100):

    std = np.std(map)
    mean = np.mean(map)
    indices = np.where(np.abs(map)>sigma*std+mean)
    map[indices] = 0.

    if inpaint_flag == True:

        mask = np.ones(map.shape)
        mask[indices] = 0.
        map = inpaint.diffusive_inpaint(map,mask,n_inpaint)

    return map

def mask_sigma_freq(tmap,sigma,inpaint_flag=False):

    ret = np.zeros(tmap.shape)

    for i in range(0,tmap.shape[2]):

        ret[:,:,i] = mask_sigma(tmap[:,:,i],sigma=sigma,inpaint_flag=inpaint_flag)

    return ret

def extract_value_at_catalogue(tsz,catalogue_tile,pix):

    i,j = get_ij_from_theta(catalogue_tile.theta_x,catalogue_tile.theta_y,pix)

    return tsz[i,j]

def get_binned_mean_map(map,pix,bins,fft_flag=False):

    map_real = rmap(pix,map)

    if fft_flag == False:

        map_fft = map_real.get_fft()

    else:

        map_fft = map

    ell = map_real.get_ell()

    binned_map = np.zeros(len(bins)-1)
    bins_centres = np.zeros(len(binned_map))

    for i in range(0,len(bins_centres)):

        indices = np.where((ell > bins[i]) & (ell < bins[i+1]))
        selected_map = np.mean(map_fft[indices]).real
        binned_map[i] = selected_map
        bins_centres[i] = 0.5*(bins[i] + bins[i+1])

    return bins_centres,binned_map

def convolve_tmap_experiment(pix,tmap,exp,freqs=None,beam_type="gaussian"):

    n_freqs = tmap.shape[2]
    freqs = np.arange(n_freqs)
    tmap_convolved = np.zeros(tmap.shape)

    for i in range(0,n_freqs):

        if beam_type == "gaussian":

            tmap_convolved[:,:,i] = get_gaussian_convolution(tmap[:,:,i],exp.FWHM[freqs[i]],pix)

        elif beam_type == "real":

            tmap_convolved[:,:,i] = get_convolution_isotropic(tmap[:,:,i],exp.get_beam(freqs[i]),pix)

    return tmap_convolved

def convolve_tmap_fft_experiment(pix,tmap_fft,exp,freqs=None,beam_type="gaussian"):

    n_freqs = tmap_fft.shape[2]
    freqs = np.arange(n_freqs)
    tmap_convolved_fft = np.zeros(tmap_fft.shape)

    for i in range(0,n_freqs):

        if beam_type == "gaussian":

            tmap_convolved_fft[:,:,i] = get_gaussian_convolution_fft(tmap_fft[:,:,i],exp.FWHM[freqs[i]],pix)

        elif beam_type == "real":

            tmap_convolved_fft[:,:,i] = get_convolution_isotropic_fft(tmap_fft[:,:,i],exp.get_beam(freqs[i]),pix)

    return tmap_convolved_fft


def get_tmap_times_fvec(tmap,fvec):

    tmap_ret = np.zeros(tmap.shape,dtype=complex)

    for i in range(0,tmap.shape[2]):

        tmap_ret[:,:,i] = tmap[:,:,i]*fvec[i]

    return tmap_ret

def get_tmap_from_map(map,n_freq):

    tmap = np.zeros((map.shape[0],map.shape[1],n_freq))

    for i in range(0,n_freq):

        tmap[:,:,i] = map

    return tmap


def select_freqs(tmap,freqs):

    return tmap[:,:,freqs]

def get_tmap_with_sed(pix,tmap,base_freq,exp,sed,beam_type="gaussian"):

    sed = sed/sed[base_freq]
    ret = np.zeros(tmap.shape)

    deconv_freq = base_freq

    if beam_type == "gaussian":

        spatial_map = get_gaussian_deconvolution(tmap[:,:,deconv_freq],exp.FWHM[deconv_freq],pix)

    elif beam_type == "real":

        spatial_map = get_deconvolution_isotropic(tmap[:,:,deconv_freq],exp.get_beam(deconv_freq),pix)

    for i in range(0,tmap.shape[2]):

        ret[:,:,i] = spatial_map*sed[i]

    ret = convolve_tmap_experiment(pix,ret,exp,freqs=np.arange(len(sed)),beam_type=beam_type)

    return ret

def get_hankel_transform(theta_range,function,n=512,pad=128):

    lrange = [1/theta_range[1], 1/theta_range[0]]

    logl1, logl2 = np.log(lrange)
    logl0        = (logl2+logl1)/2
    dlog    = (logl2-logl1)/n
    i0           = (n+1)/2+pad
    l       = np.exp(logl0 + (np.arange(1,n+2*pad+1)-i0)*dlog)
    r       = 1/l[::-1]
    rprof = function(r)
    lprof = 2*np.pi*scipy.fft.fht(rprof*r,dlog,0)/l

    res = [l,lprof]
    l,lprof = tuple([arr[...,pad:-pad] for arr in res])

    return l,r,lprof,dlog

class RadialFourierTransform:

    def __init__(self, lrange=None, rrange=None, n=512//4, pad=256//4):
        """Construct an object for transforming between radially
        symmetric profiles in real-space and fourier space using a
        fast Hankel transform. Aside from being fast, this is also
        good for representing both cuspy and very extended profiles
        due to the logarithmically spaced sample points the fast
        Hankel transform uses. A cost of this is that the user can't
        freely choose the sample points. Instead one passes the
        multipole range or radial range of interest as well as the
        number of points to use.

        The function currently assumes two dimensions with flat geometry.
        That means the function is only approximate for spherical
        geometries, and will only be accurate up to a few degrees
        in these cases.

        Arguments:
        * lrange = [lmin, lmax]: The multipole range to use. Defaults
          to [0.01, 1e6] if no rrange is given.
        * rrange = [rmin, rmax]: The radius range to use if lrange is
        	not specified, in radians. Example values: [1e-7,10].
        	Since we don't use spherical geometry r is not limited to 2 pi.
        * n: The number of logarithmically equi-spaced points to use
        	in the given range. Default: 512. The Hankel transform usually
        	doesn't need many points for good accuracy, and can suffer if
        	too many points are used.
        * pad: How many extra points to pad by on each side of the range.
          Padding is useful to get good accuracy in a Hankel transform.
          The transforms this function does will return padded output,
        	which can be unpadded using the unpad method. Default: 256
        """
        if lrange is None and rrange is None: lrange = [0.1, 1e7]
        if lrange is None: lrange = [1/rrange[1], 1/rrange[0]]
        logl1, logl2 = np.log(lrange)
        logl0        = (logl2+logl1)/2
        self.dlog    = (logl2-logl1)/n
        i0           = (n+1)/2+pad
        self.l       = np.exp(logl0 + (np.arange(1,n+2*pad+1)-i0)*self.dlog)
        self.r       = 1/self.l[::-1]
        self.pad     = pad

    def real2harm(self, rprof):
        """Perform a forward (real -> harmonic) transform, taking us from the
        provided real-space radial profile rprof(r) to a harmonic-space profile
        lprof(l). rprof can take two forms:
        1. A function rprof(r) that can be called to evalute the profile at
           arbitrary points.
        2. An array rprof[self.r] that provides the profile evaluated at the
           points given by this object's .r member.
        The transform is done along the last axis of the profile.
        Returns lprof[self.l]. This includes padding, which can be removed
        using self.unpad"""
        import scipy.fft
        try: rprof = rprof(self.r)
        except TypeError: pass
        lprof = 2*np.pi*scipy.fft.fht(rprof*self.r, self.dlog, 0)/self.l
        return lprof

    def harm2real(self, lprof):
        """Perform a backward (harmonic -> real) transform, taking us from the
        provided harmonic-space radial profile lprof(l) to a real-space profile
        rprof(r). lprof can take two forms:
        1. A function lprof(l) that can be called to evalute the profile at
           arbitrary points.
        2. An array lprof[self.l] that provides the profile evaluated at the
           points given by this object's .l member.
        The transform is done along the last axis of the profile.
        Returns rprof[self.r]. This includes padding, which can be removed
        using self.unpad"""
        import scipy.fft
        try: lprof = lprof(self.l)
        except TypeError: pass
        rprof = scipy.fft.ifht(lprof/(2*np.pi)*self.l, self.dlog, 0)/self.r
        return rprof

    def unpad(self, *arrs):
        """Remove the padding from arrays used by this object. The
        values in the padded areas of the output of the transform have
        unreliable values, but they're not cropped automatically to
        allow for round-trip transforms. Example:
        	r = unpad(r_padded)
        	r, l, vals = unpad(r_padded, l_padded, vals_padded)"""
        if self.pad == 0: res = arrs
        else: res = tuple([arr[...,self.pad:-self.pad] for arr in arrs])
        return res[0] if len(arrs) == 1 else res


def profile_to_tform_hankel(profile_fun, lmin=0.1, lmax=1e7, n=512, pad=256):
    """Transform a radial profile given by the function profile_fun(r) to
    sperical harmonic coefficients b(l) using a Hankel transform. This approach
    is good at handling cuspy distributions due to using logarithmically spaced
    points. n points from 10**logrange[0] to 10**logrange[1] will be used.
    Returns l, bl. l will not be equi-spaced, so you may want to interpolate
    the results. Note that unlike other similar functions in this module and
    the curvedsky module, this function uses the flat sky approximation, so
    it should only be used for profiles up to a few degrees in size."""
    rht   = RadialFourierTransform(lrange=[lmin,lmax], n=n, pad=pad)
    lprof = rht.real2harm(profile_fun)
    return rht.unpad(rht.l, lprof)


def bin_radially(pix,map,bins_edges=None,nbins=None,theta_misc=[0.,0.]):

    theta_map = rmap(pix).get_distance_map_wrt_centre(theta_misc)

    if bins_edges is None:

        bins_edges = np.linspace(0,np.max(theta_map),nbins)

    bins = np.zeros(len(bins_edges)-1)
    binned_map = np.zeros(len(bins_edges)-1)

    theta_flattened = theta_map.flatten()
    map_flattened = map.flatten()

    for i in range(0,len(bins_edges)-1):

        bins[i] = 0.5*(bins_edges[i+1]+bins_edges[i])
        indices = np.where((theta_flattened < bins_edges[i+1]) & (theta_flattened >= bins_edges[i]))[0]
        binned_map[i] = np.mean(map_flattened[indices])

    return bins,binned_map



import numpy as np
from astropy.cosmology import Planck15
from scipy import integrate
import spec
import maps
import model
import mmf
import expt
import time
import healpy as hp

def get_realisation_from_spec_fft(pix,ell,cl):

    my_map = maps.rmap(pix)

    nx = pix.nx
    ny = pix.ny
    map_fft = (np.random.standard_normal((nx,ny//2+1))+1.j*np.random.standard_normal((nx,ny//2+1)))/np.sqrt(2.)
    map_fft[0,0] = np.sqrt(2.)*np.real(map_fft[0,0])
    map_fft[nx//2+1:,0] = np.conj(map_fft[1:nx//2,0][::-1])

    map_fft = maps.rfft2_to_fft2(pix,map_fft)
    map_fft[:,:] *= np.interp(my_map.get_ell(),ell,np.sqrt(cl),right=0)

    return map_fft

def get_realisation_from_spec_real(pix,ell,cl):

    return maps.get_ifft(get_realisation_from_spec_fft(pix,ell,cl),pix)


def get_cmb_realisation(pix): #CMB realisation in muK

    ell,cl = spec.get_camb_cltt()

    return get_realisation_from_spec_real(pix,ell,cl).real

#Returns CMB realisation convolved with Gaussian Planck beams in muK (all channels)

def get_cmb_realisation_gaussian(pix,n_freq=None,non_periodic=True,exp=None,beam="gaussian"):

    n_freq = exp.n_freqs

    if non_periodic == True:

        fac = 2
        nx = pix.nx*fac
        ny = pix.ny*fac
        pix = maps.pixel(nx,pix.dx,ny=ny)

    cmb_realisation = np.zeros((pix.nx,pix.ny,n_freq))

    cmb = get_cmb_realisation(pix)

    for i in range(0,n_freq):

        if beam == "gaussian":

            cmb_realisation[:,:,i] = maps.get_gaussian_convolution(cmb,exp.FWHM[i],pix)

        elif beam == "real":

            cmb_realisation[:,:,i] = maps.get_convolution_isotropic(cmb,exp.get_beam(i),pix)


    if non_periodic == True:

        cmb_realisation = cmb_realisation[nx-nx//fac:nx+nx//fac,ny-ny//fac:ny+ny//fac,:]

    return cmb_realisation



def get_gaussian_realisation(pix,n_freq=None,non_periodic=True,exp=None,beam="gaussian",field="cmb",z=0.):

    n_freq = exp.n_freqs

    if non_periodic == True:

        fac = 2
        nx = pix.nx*fac
        ny = pix.ny*fac
        pix = maps.pixel(nx,pix.dx,ny=ny)

    realisation = np.zeros((pix.nx,pix.ny,n_freq))

    if field == "cmb":

        ell,cl = spec.get_camb_cltt()
        cl_nu = np.repeat(cl[:,np.newaxis],2,axis=1)

    elif field == "cib":

        ell = np.arange(2,100000)
        cross_spec = spec.cross_spec_theory(ell,exp,cmb_flag=False,cib_flag=True,tsz_flag=False,
        noise_flag=False,tsz_cib_flag=False)

        cl_nu = cross_spec[:,i]


    realisation_unconvolved = get_realisation_from_spec_real(pix,ell,cl).real

    for i in range(0,n_freq):

        if beam == "gaussian":

            realisation[:,:,i] = maps.get_gaussian_convolution(cmb,exp.FWHM[i],pix)*sed[i]

        elif beam == "real":

            realisation[:,:,i] = maps.get_convolution_isotropic(cmb,exp.get_beam(i),pix)*sed[i]


    if non_periodic == True:

        cmb_realisation = cmb_realisation[nx-nx//fac:nx+nx//fac,ny-ny//fac:ny+ny//fac,:]

    return cmb_realisation

class simulated_map:

    def __init__(self,pix,pixel_id=0,cosmology=Planck15,n_clus=None,
    generate=True,non_periodic=True,cmb_flag=True,noise_flag=True,cluster_flag=True,
    catalogue=None,concentration=1.177,beam="gaussian",
    locations_clusters="random",pixel_buffer_cluster=0.,n_freqs=6,exp=None):

        t_obs = None
        cmb = None
        noise = None
        cluster_sz = None

        if generate == True:

            t_obs = np.zeros((pix.nx,pix.ny,n_freqs))

            if noise_flag == True:

                noise = maps.get_noise_tmap_exp(pix,exp)
                t_obs += noise

            if cmb_flag == True:

                cmb = get_cmb_realisation_gaussian(pix,non_periodic=non_periodic,exp=exp,beam=beam)
                t_obs += cmb

            if cluster_flag == True:

                if catalogue is not None:

                    cluster_tsz,catalogue = get_cluster_tsz_map(pix,catalogue,pixel_id=pixel_id,cosmology=cosmology,
                    concentration=concentration,pixel_buffer=pixel_buffer_cluster,locations=locations_clusters,beam=beam)
                    t_obs += cluster_tsz

        self.t_obs = t_obs
        self.cmb = cmb
        self.noise = noise
        self.cluster_sz = cluster_sz
        self.true_catalogue = catalogue
        self.n_freqs = 6.

    def save(self,name):

        np.save(name + "_map.npy",(self.t_obs,self.cmb,self.noise,self.cluster_sz,self.true_catalogue))

    def load(self,name):

        self.t_obs,self.cmb,self.noise,self.cluster_sz,self.true_catalogue = np.load(name + "_map.npy",allow_pickle=True)


def get_cluster_tsz_map(pix,catalogue,nfreq=6,cosmology=Planck15,concentration=1.177,pixel_buffer=0.,locations="random",pixel_id=0,beam="gaussian"):

    m_500_vec = catalogue.m_500
    z_vec = catalogue.z
    n_clus = len(m_500_vec)

    t_tsz = np.zeros((pix.nx,pix.ny,nfreq))

    if n_clus > 0:

        theta_500_vec = model.get_theta_500_arcmin(m_500_vec,z_vec,cosmology)
        pixel_ids = np.ones(n_clus)*pixel_id

        y0_vec = np.zeros(n_clus)

        if locations == "random":

            theta_x_vec = np.random.rand(n_clus)*(pix.nx-pixel_buffer)*pix.dx+pixel_buffer*pix.dx
            theta_y_vec = np.random.rand(n_clus)*(pix.nx-pixel_buffer)*pix.dx+pixel_buffer*pix.dx

            for i in range(0,n_clus):

                theta_misc = maps.get_theta_misc([theta_x_vec[i],theta_y_vec[i]],pix)
                nfw = model.gnfw_arnaud(model.get_m_500(theta_500_vec[i],z_vec[i],cosmology),z_vec[i],cosmology)
                t_cluster = nfw.get_t_map_convolved(pix,"Planck",theta_misc=theta_misc,beam=beam)
                t_tsz += t_cluster
                y0_vec[i] = nfw.get_y_norm()

        catalogue = mmf.cluster_catalogue(y0=y0_vec,
        theta_500=theta_500_vec,theta_x=theta_x_vec,theta_y=theta_y_vec,pixel_ids=pixel_ids)

    return t_tsz,catalogue

def get_cluster_tsz_map_from_cat(pix,catalogue,nfreq=6,cosmology=Planck15,concentration=1.177,beam="gaussian"):

    m_500_vec = catalogue.m_500
    z_vec = catalogue.z
    n_clus = len(m_500_vec)

    t_tsz = np.zeros((pix.nx,pix.ny,nfreq))

    if n_clus > 0:

        y0_vec = np.zeros(n_clus)

        for i in range(0,n_clus):

            theta_misc = maps.get_theta_misc([catalogue.theta_x[i],catalogue.theta_y[i]],pix)
            nfw = model.gnfw_arnaud(m_500_vec[i]/1e15,z_vec[i],cosmology)
            t_cluster = nfw.get_t_map_convolved(pix,"Planck",theta_misc=theta_misc,beam=beam)
            t_tsz += t_cluster
            y0_vec[i] = nfw.get_y_norm()

    catalogue.y0 = y0_vec

    return t_tsz

def sample_on_sphere(n_samples):

    vec = np.random.randn(3,n_samples)
    vec /= np.linalg.norm(vec, axis=0)
    lon,lat = hp.pixelfunc.vec2ang(vec,lonlat=True)

    return lon,lat

def get_white_noise_hpmap(nside,n_lev):

    pixel_area = hp.pixelfunc.nside2pixarea(nside)
    sigma_pixel = n_lev/np.sqrt(pixel_area)/(180.*60./np.pi)
    n_pixels = hp.pixelfunc.nside2npix(nside)

    return np.random.standard_normal(n_pixels)*sigma_pixel

"""
class sim_full_sky_maps:

    def __init__(nside=4096,exp="Planck",freqs=[0,1,2,3,4,5],cmb_flag=True,cib_flag=True,
    tsz_flag=True,ksz_flag=True,gdust_flag=True,synchro_flag=True,freefree_flag=True,radio_flag=True):

        npix = hp.nside2npix(nside)
        map = np.zeros(npix)

        if cib_flag == True:
"""



import numpy as np
from astropy.cosmology import Planck15
from scipy import integrate
import maps
import expt
import scipy.optimize as optimize
import pylab as pl
import websky
import scipy
#from pysz import pysz

class gnfw_arnaud:

    def __init__(self,M_500,z_halo,cosmology,c_500=1.177,path="/Users/user/Desktop/"):

        #Create variables

        self.M_500 = M_500*1e15
        self.c = c_500
        self.z_halo = z_halo
        self.cosmology = cosmology
        self.path = path

        #Calculate basic quantities

        const = constants()
        self.rho_c = cosmology.critical_density(self.z_halo).value*1000.*const.mpc**3/const.solar
        self.r_s = (3./4.*self.M_500/self.rho_c/500./np.pi/self.c**3)**(1./3.)
        self.R_500 = self.c*self.r_s

        self.chi_h = cosmology.comoving_distance(self.z_halo).value #In Mpc
        self.d_ad_h = self.chi_h/(1.+z_halo)

        self.theta_500 = self.R_500/self.d_ad_h
        self.theta_500_arcmin = self.theta_500/np.pi*180.*60.
        self.my_params_sz = params_sz(cosmology,c=self.c)

        self.Ez = self.cosmology.H(self.z_halo).value/self.cosmology.H0.value
        self.h70 = self.cosmology.H0.value/70.
        self.P_500 = 1.65e-3*self.Ez**(8./3.)*(self.M_500/3e14*self.h70)**(2./3.)*self.h70**2  #in units of keV cm-3
        const = constants()
        self.P2Y = const.mpc*100.*const.sigma_thomson_e/const.mass_e #converts pressure in units of keV cm-3 to dimensionless Compton y
        self.Y_500 = self.P_500*self.R_500**3*4.*np.pi/3.*self.P2Y #characteristic Y, in units of Mpc^2


    def get_p_cal_map(self,pix,c_500,theta_misc=[0.,0.]):

        nx = pix.nx
        dx = pix.dx
        ny = pix.ny
        dy = pix.dy

        theta_max = np.sqrt((nx*dx)**2+(ny*dy)**2)*0.5*1.1
        theta_vec = np.linspace(0.0001,theta_max,1000)
        p_cal_vec = np.zeros(len(theta_vec))

        for i in range(0,len(theta_vec)):

            x = theta_vec[i]/self.theta_500
            p_cal_vec[i] = self.get_p_cal_int(self.my_params_sz,x)

        theta_map = maps.rmap(pix).get_distance_map_wrt_centre(theta_misc)
        p_cal_map = np.interp(theta_map,theta_vec,p_cal_vec)

        return p_cal_map

    def get_y_map(self,pix,theta_misc=[0.,0.]):

        p_cal_map = self.get_p_cal_map(pix,self.c,theta_misc)
        y_map = self.p_cal_to_y(p_cal_map)

        return y_map

    def get_y_norm(self,type="R_500"):  #returns y at R_500

        if type == "R_500":

            r = 1.

        elif type == "centre":

            r = 1e-6

        return self.p_cal_to_y(self.get_p_cal_int(self.my_params_sz,r))

    def get_y_at_angle(self,theta): #angle in rad

        return self.p_cal_to_y(self.get_p_cal_int(self.my_params_sz,theta/self.theta_500))

    def get_y_map_convolved(self,pix,fwhm_arcmin,theta_misc=[0.,0.]):

        return maps.get_gaussian_convolution(self.get_y_map(pix,theta_misc=theta_misc),fwhm_arcmin,pix)

    def get_y_norm_convolved(self,pix,fwhm_arcmin,y_map=None,theta_misc=[0.,0.]):

        if y_map == None:

            y_map = self.get_y_map_convolved(pix,fwhm_arcmin,theta_misc)

        (theta_x,theta_y) = theta_misc
        x_coord = maps.rmap(pix).get_x_coord_map_wrt_centre(theta_x)[0,:]/self.theta_500
        y_coord = maps.rmap(pix).get_y_coord_map_wrt_centre(theta_y)[:,0]/self.theta_500

        return interpolate.interp2d(x_coord,y_coord,y_map)(1.,0.)


    def p_cal_to_y(self,p_cal_map):

        prefactor = self.P_500*(self.M_500/3e14*self.h70)**(self.my_params_sz.alpha_p)*self.R_500

        return p_cal_map*prefactor*self.P2Y

    def get_Y_sph(self,x):  #returns spherically integrated Y, assuming self-similarity+alpha_p, in units of Mpc^2. x in units of R_500

        if x == 1:

            I = 0.6145

        elif x == 5:

            I = 1.1037

        return self.Y_500*(self.M_500/3e14*self.h70)**(self.my_params_sz.alpha_p)*I

    def get_Y_cyl(self,x):  #returns cylindrically integrated Y, assuming self-similarity+alpha_p. x in units of R_500.
                            #In units of Mpc^2, i.e., divide by d_A^2 to get true aperture integrated Y

        if x == 1:

            J = 0.7398

        elif x == 5:

            J = 1.1037

        return self.Y_500*(self.M_500/3e14*self.h70)**(self.my_params_sz.alpha_p)*J

    def get_Y_aperture(self,x): #dimensionless (sterad)

        return self.get_Y_cyl(x)/self.d_ad_h**2

    def get_p_cal_int(self,my_params_sz,r): #r is distance in units of R_500

        def p_cal(z):

            x = np.sqrt(z**2+r**2)

            if x < 5:  #truncating at 5 R_500

                p_cal = my_params_sz.P0/((my_params_sz.c_500*x)**my_params_sz.gamma*(1.+(my_params_sz.c_500*x)**my_params_sz.alpha)**((my_params_sz.beta-my_params_sz.gamma)/my_params_sz.alpha))


            else:

                p_cal = 0.

            return p_cal

        #p_int = integrate.quad(p_cal,-np.inf,np.inf,epsabs=1.49e-08/100.)[0]
        p_int = integrate.quad(p_cal,-5.,5.,epsabs=1.49e-08/100.,limit=100)[0]

        return p_int

    def get_t_map(self,pix,exp,theta_misc=[0.,0.],sed=None): #returns t map in units of muK (all freqs)

        y_map = self.get_y_map(pix,theta_misc=theta_misc)

        if sed == None:

            sed = exp.tsz_f_nu

        n_freqs = exp.n_freqs
        t_map = np.zeros((pix.nx,pix.ny,n_freqs))

        for i in range(0,n_freqs):

            t_map[:,:,i] = y_map*sed[i]

        return t_map

    #NOTE: theta_misc is in rad and is [i,j] coord. Positive miscentering in i
    #means "upwards", and in j "leftwards" (i.e., in both cases towards the origin)

    def get_t_map_convolved(self,pix,exp,theta_misc=[0.,0.],theta_cart=None,beam="gaussian",get_nc=False):

        if theta_cart != None:

            theta_misc = maps.get_theta_misc(theta_cart,pix)

        tmap = self.get_t_map(pix,exp,theta_misc=theta_misc)
        tmap_convolved = maps.convolve_tmap_experiment(pix,tmap,exp,beam_type=beam)
        #tmap_convolved = tmap

        if get_nc == True:

            ret = tmap_convolved,tmap

        else:

            ret = tmap_convolved

        return ret

#Delta critical always

class gnfw_tsz:

    def __init__(self,M,z_halo,cosmology,c=1.177,Delta=500.,type="arnaud",path="/Users/user/Desktop/"):

        #Create variables

        self.M = M*1e15
        self.c = c
        self.z_halo = z_halo
        self.cosmology = cosmology
        self.Delta = Delta
        self.type = type
        self.path = path

        #Calculate basic quantities

        const = constants()
        self.rho_c = cosmology.critical_density(self.z_halo).value*1000.*const.mpc**3/const.solar
        self.R_Delta = (3./4.*self.M/self.rho_c/self.Delta/np.pi)**(1./3.)

        self.chi = cosmology.comoving_distance(self.z_halo).value #In Mpc
        self.d_ad = self.chi/(1.+z_halo)

        self.theta_Delta = self.R_Delta/self.d_ad
        self.theta_Delta_arcmin = self.theta_Delta/np.pi*180.*60.
        self.my_params_sz = params_sz(cosmology,c=self.c,type=self.type,M=self.M,z=self.z_halo)

        const = constants()
        self.P2Y = const.mpc*100.*const.sigma_thomson_e/const.mass_e #converts pressure in units of keV cm-3 to dimensionless Compton y

        if self.type == "arnaud":

            self.Ez = self.cosmology.H(self.z_halo).value/self.cosmology.H0.value
            self.h70 = self.cosmology.H0.value/70.
            self.P_500 = 1.65e-3*self.Ez**(8./3.)*(self.M/3e14*self.h70)**(2./3.)*self.h70**2  #in units of keV cm-3
            self.Y_500 = self.P_500*self.R_Delta**3*4.*np.pi/3.*self.P2Y #characteristic Y, in units of Mpc^2

        #elif self.type == "battaglia":

        #    self.rho_c = cosmology.critical_density(self.z_halo).value*1000.*const.mpc**3/const.solar
    #          self.P_Delta = self.G*self.M*self.Delta*self.rho_c*self.cosmology.Omb/self.cosmology.Om0/(2.*self.R_Delta)


    def get_p_cal_map(self,pix,c_500,theta_misc=[0.,0.]):

        #theta_max = np.sqrt((nx*dx)**2+(ny*dy)**2)*0.5*1.1
        theta_max = self.theta_Delta*self.my_params_sz.R_truncation
        theta_vec = np.linspace(0.,theta_max,1000)
        p_cal_vec = np.zeros(len(theta_vec))

        for i in range(0,len(theta_vec)):

            x = theta_vec[i]/self.theta_Delta
            p_cal_vec[i] = self.get_p_cal_int(x)

        theta_map = maps.rmap(pix).get_distance_map_wrt_centre(theta_misc)
        p_cal_map = np.interp(theta_map,theta_vec,p_cal_vec,right=0.)

        return p_cal_map

    def get_y_map(self,pix,theta_misc=[0.,0.]):

        p_cal_map = self.get_p_cal_map(pix,self.c,theta_misc)
        y_map = self.p_cal_to_y(p_cal_map)

        return y_map

    def get_y_norm(self,type="centre"):

        if type == "R_500":

            r = 1.

        elif type == "centre":

            r = 1e-6

        return self.p_cal_to_y(self.get_p_cal_int(r))

    def get_y_at_angle(self,theta): #angle in rad

        return self.p_cal_to_y(self.get_p_cal_int(theta/self.theta_Delta))

    def get_y_map_convolved(self,pix,fwhm_arcmin,theta_misc=[0.,0.]):

        return maps.get_gaussian_convolution(self.get_y_map(pix,theta_misc=theta_misc),fwhm_arcmin,pix)

    def get_y_norm_convolved(self,pix,fwhm_arcmin,y_map=None,theta_misc=[0.,0.]):

        if y_map == None:

            y_map = self.get_y_map_convolved(pix,fwhm_arcmin,theta_misc)

        (theta_x,theta_y) = theta_misc
        x_coord = maps.rmap(pix).get_x_coord_map_wrt_centre(theta_x)[0,:]/self.theta_Delta
        y_coord = maps.rmap(pix).get_y_coord_map_wrt_centre(theta_y)[:,0]/self.theta_Delta

        return interpolate.interp2d(x_coord,y_coord,y_map)(1.,0.)


    def p_cal_to_y(self,p_cal_map):

        if self.type == "arnaud":

            prefactor = self.P_500*(self.M/3e14*self.h70)**(self.my_params_sz.alpha_p)*self.R_Delta

        elif self.type == "battaglia":

            prefactor = 2.61051e-18*self.cosmology.Ob0/self.cosmology.Om0*self.M*self.cosmology.H(self.z_halo).value**2/self.R_Delta*self.R_Delta/1e3

        return p_cal_map*prefactor*self.P2Y

    def get_Y_sph(self,x):  #returns spherically integrated Y, assuming self-similarity+alpha_p, in units of Mpc^2. x in units of R_500

        if x == 1:

            I = 0.6145

        elif x == 5:

            I = 1.1037

        return self.Y_500*(self.M_500/3e14*self.h70)**(self.my_params_sz.alpha_p)*I

    def get_Y_cyl(self,x):  #returns cylindrically integrated Y, assuming self-similarity+alpha_p. x in units of R_500.
                            #In units of Mpc^2, i.e., divide by d_A^2 to get true aperture integrated Y

        if x == 1:

            J = 0.7398

        elif x == 5:

            J = 1.1037

        return self.Y_500*(self.M_500/3e14*self.h70)**(self.my_params_sz.alpha_p)*J

    def get_Y_aperture(self,x): #dimensionless (sterad)

        return self.get_Y_cyl(x)/self.d_ad_h**2

    def pressure_profile(self,x):  #x is distance in units of R_500

        if x < self.my_params_sz.R_truncation:

            if self.type == "arnaud":

                p_cal = self.my_params_sz.P0/((self.my_params_sz.c_500*x)**self.my_params_sz.gamma*(1.+(self.my_params_sz.c_500*x)**self.my_params_sz.alpha)**((self.my_params_sz.beta-self.my_params_sz.gamma)/self.my_params_sz.alpha))

            elif self.type == "battaglia":

                p_cal = self.my_params_sz.P0*(x/self.my_params_sz.xc)**self.my_params_sz.gamma*(1.+(x/self.my_params_sz.xc)**self.my_params_sz.alpha)**(-self.my_params_sz.beta)

        else:

            p_cal = 0.

        return p_cal

    def get_p_cal_int(self,r): #r is distance in units of R_500

        def p_cal(z_los):

            x = np.sqrt(z_los**2+r**2)

            return self.pressure_profile(x)

        p_int = integrate.quad(p_cal,-self.my_params_sz.R_truncation,self.my_params_sz.R_truncation,
        epsabs=1.49e-08/100.,limit=100)[0]

        return p_int

    def get_t_map(self,pix,exp,theta_misc=[0.,0.],eval_type="standard",sed=None): #returns t map in units of muK

        if eval_type == "standard":

            y_map = self.get_y_map(pix,theta_misc=theta_misc)

        elif eval_type == "hankel":

            y_map = self.get_y_map_hankel(pix,theta_misc=theta_misc)

        n_freqs = exp.n_freqs

        if sed is None:

            sed = exp.tsz_f_nu

        elif sed is False:

            sed = np.ones(n_freqs)

        t_map = np.zeros((pix.nx,pix.ny,n_freqs))

        for i in range(0,n_freqs):

            t_map[:,:,i] = y_map*sed[i]

        return t_map

    def get_y_map_hankel(self,pix,theta_misc=[0.,0.]):

        def to_transform(theta):

            return np.vectorize(self.get_p_cal_int)(theta/self.theta_Delta)

        theta_range = [pix.dx/20.,self.theta_Delta*self.my_params_sz.R_truncation*10.]

        rht = maps.RadialFourierTransform()
        rprofs       = to_transform(rht.r)
        lprofs       = rht.real2harm(rprofs)
        ell_vec = rht.l
        rprofs     = rht.harm2real(lprofs)
        r, rprofs    = rht.unpad(rht.r, rprofs)

        theta_map = maps.rmap(pix).get_distance_map_wrt_centre(theta_misc)
        p_cal_map = np.interp(theta_map,r,rprofs,right=0.)
        y_map = self.p_cal_to_y(p_cal_map)

        return y_map

    def get_t_map_convolved_hankel(self,pix,exp,theta_misc=[0.,0.],beam_type="gaussian",get_nc=False,sed=None):

        if sed is None:

            sed = exp.tsz_f_nu

        elif sed is False:

            sed = np.ones(len(exp.tsz_f_nu))

        def to_transform(theta):

            #return np.vectorize(self.pressure_profile)(theta/self.theta_Delta)

            return np.vectorize(self.get_p_cal_int)(theta/self.theta_Delta)

        theta_range = [pix.dx/20.,self.theta_Delta*self.my_params_sz.R_truncation*10.]

        rht = maps.RadialFourierTransform()
        rprofs       = to_transform(rht.r)
        lprofs       = rht.real2harm(rprofs)
        ell_vec = rht.l

        t_map = np.zeros((pix.nx,pix.ny,exp.n_freqs))
        t_map_conv = np.zeros((pix.nx,pix.ny,exp.n_freqs))

        for i in range(0,exp.n_freqs):

            if beam_type == "gaussian":

                beam_fft = maps.get_bl(exp.FWHM[i],ell_vec)

            elif beam_type == "real":

                ell_beam,beam_fft = exp.get_beam(i)
                beam_fft = np.interp(ell_vec,ell_beam,beam_fft)

            rprofs_convolved      = rht.harm2real(lprofs*beam_fft)
            r, rprofs_convolved    = rht.unpad(rht.r, rprofs_convolved)

            rprofs      = rht.harm2real(lprofs)
            r, rprofs    = rht.unpad(rht.r, rprofs)

            theta_map = maps.rmap(pix).get_distance_map_wrt_centre(theta_misc)

            p_cal_map_conv = np.interp(theta_map,r,rprofs_convolved,right=0.)
            p_cal_map = np.interp(theta_map,r,rprofs,right=0.)

            t_map_conv[:,:,i] = self.p_cal_to_y(p_cal_map_conv)*sed[i]
            t_map[:,:,i] = self.p_cal_to_y(p_cal_map)*sed[i]

        if get_nc == True:

            ret = t_map_conv,t_map

        else:

            ret = t_map_conv

        return ret

    #NOTE: theta_misc is in rad and is [i,j] coord. Positive miscentering in i
    #means "upwards", and in j "leftwards" (i.e., in both cases towards the origin)

    def get_t_map_convolved(self,pix,exp,theta_misc=[0.,0.],theta_cart=None,beam="gaussian",get_nc=False,
    eval_type="standard",sed=None):

        if theta_cart != None:

            theta_misc = maps.get_theta_misc(theta_cart,pix)

        if eval_type == "standard":

            tmap = self.get_t_map(pix,exp,theta_misc=theta_misc,eval_type=eval_type,sed=sed)
            tmap_convolved = maps.convolve_tmap_experiment(pix,tmap,exp,beam_type=beam)

        elif eval_type == "hankel":

            tmap_convolved,tmap = self.get_t_map_convolved_hankel(pix,exp,theta_misc=theta_misc,beam_type=beam,get_nc=True,sed=sed)

        if get_nc == True:

            ret = tmap_convolved,tmap

        else:

            ret = tmap_convolved

        return ret

class params_sz: #For above class

    def __init__(self,cosmology,c,z=None,M=None,type="arnaud"):

        if type == "arnaud":

            self.P0 = 8.403*(cosmology.H0.value/70.)**(-1.5)
            self.c_500 = c #1.77 from Arnaud
            self.gamma = 0.3081
            self.alpha = 1.0510
            self.beta = 5.4905
            self.alpha_p = 0.12
            self.R_truncation = 5. #in units of R_500

        elif type == "battaglia":

            self.P0 = param_scal_rel(M,z,18.1,0.154,-0.758)
            self.xc = param_scal_rel(M,z,0.497,-0.00865,0.731)
            self.alpha = 1.
            self.beta = param_scal_rel(M,z,4.35,0.0393,0.415)
            self.gamma = -0.3
            self.R_truncation = 4. #in units of R_200

def param_scal_rel(M,z,A0,alpha_m,alpha_z):

    return A0*(M/1e14)**alpha_m*(1.+z)**alpha_z



class constants:

    def __init__(self):

        self.c_light = 299792458. #in m/s
        self.G = 6.674e-11
        self.solar = 1.98855e30    #Solar mass in kg
        self.mpc = 3.08567758149137e22   #Mpc in m
        self.sigma_thomson_e = 6.6524587158e-25 #in cm2
        self.mass_e = 0.51099895e3 # in keV
        self.k_B = 1.38064852e-23 #in J/K
        self.T_CMB = 2.7255
        self.h = 6.62607004e-34 #in Js



def get_m_500(theta_500_arcmin,z,cosmology): #return M_500 in units of 1e15 solar masses

    theta_500 = theta_500_arcmin/60./180.*np.pi
    R_500 = theta_500*cosmology.comoving_distance(z).value/(1+z)
    const = constants()
    rho_c = cosmology.critical_density(z).value*1000.*const.mpc**3/const.solar
    M_500 = 500.*4.*np.pi/3.*rho_c*R_500**3/1e15

    return M_500

def get_theta_500_arcmin(M_500,z,cosmology): #return M_500 in units of 1e15 solar masses

    M_500 *= 1e15
    const = constants()
    rho_c = cosmology.critical_density(z).value*1000.*const.mpc**3/const.solar
    R_500 = (M_500/(500.*4.*np.pi/3.*rho_c))**(1./3.)
    theta_500 = R_500/(cosmology.comoving_distance(z).value/(1+z))
    theta_500_arcmin = theta_500*60.*180./np.pi

    return theta_500_arcmin

def get_tsz_f_nu(nu,units): #nu in Hz, non relativistic, returns tSZ spectral signature, which times Compton y gives frequency map

    const = constants()
    x = nu*const.h/(const.k_B*const.T_CMB)
    ret = x*(np.exp(x)+1.)/(np.exp(x)-1.) - 4.

    if units == "TCMB":

        ret = ret

    elif units == "muK":

        ret *= const.T_CMB*1e6

    elif units == "K":

        ret *= const.T_CMB

    return ret

def get_t_map(y0,theta_500,theta_misc,cosmology,pix,exp):

    z = 0.2
    M_500 = get_m_500(theta_500,z,cosmology)
    nfw = gnfw_arnaud(M_500,z,cosmology)
    t_tem = nfw.get_t_map_convolved_gaussian(pix,exp,theta_misc=theta_misc)/nfw.get_y_norm()*y0

    return t_tem

def get_t_map_from_catalogue(pix,catalogue,z,cosmology=Planck15,exp=None):

    if exp is None:

        exp = expt.planck_specs()

    catalogue.purge_minus_1()
    n_clus = len(catalogue.q_opt)

    if isinstance(z,float):

        z = np.ones(n_clus)*z

    t_map = np.zeros((pix.nx,pix.ny,6))

    theta_x = catalogue.theta_x
    theta_y = catalogue.theta_y
    theta_500 = catalogue.theta_500
    y0 = catalogue.y0

    for i in range(0,n_clus):

        theta_cart = [theta_x[i],theta_y[i]]
        theta_misc = maps.get_theta_misc(theta_cart,pix)
        nfw = gnfw_arnaud(get_m_500(theta_500[i],z[i],cosmology),z[i],cosmology)
        t_cluster = nfw.get_t_map_convolved_gaussian(pix,exp,theta_misc=theta_misc)
        t_map += t_cluster

    return t_map

def g(x):

    return (np.log(1+x)-x/(1.+x))/x**3

#solve for mass 2

class mass_conversion:

    def __init__(self,cosmology,Delta1,crit1,Delta2,crit2,c1):

        self.cosmology = cosmology
        self.Delta1 = Delta1
        self.Delta2 = Delta2
        self.crit1 = crit1
        self.crit2 = crit2
        self.c1 = c1

    def get_ratio_delta(self,redshift):

        if self.crit1 == self.crit2:

            ratio_delta = np.ones(len(redshift))

        if self.crit1 == "mean" and self.crit2 == "critical":

            ratio_delta = self.cosmology.H(redshift)/(self.cosmology.Om0*(1.+redshift**3))/self.cosmology.H0

        if self.crit1 == "critical" and self.crit2 == "mean":

            ratio_delta = 1./(self.cosmology.H(redshift)/(self.cosmology.Om0*(1.+redshift**3))/self.cosmology.H0)

        ratio_delta = ratio_delta*self.Delta2/self.Delta1
        self.ratio_delta = ratio_delta

        return ratio_delta

    def get_c2(self,redshift):

        ratio_delta = self.get_ratio_delta(redshift)
        c2_vec = np.zeros(len(redshift))

        for i in range(0,len(redshift)):

            if i % 1000 == 0:

                print(i)

            def f2root(x):

                return g(x)-g(self.c1)*ratio_delta[i]

            c2_vec[i] = optimize.root_scalar(f2root,x0=5.,x1=3.).root

        self.c2 = c2_vec

        return c2_vec

    def get_m2(self,M1,redshift,R1=None): #M in units of 1e15 M_Sun

        if R1 is None:

            R1 =  get_R_Delta(M1,redshift,self.cosmology,self.Delta1,self.crit1)

        c2 = self.get_c2(redshift)
        R2 = c2/self.c1*R1
        M2 = self.ratio_delta*(c2/self.c1)**3*M1

        self.R2 = R2
        self.M2 = M2

        return (M2,R2)

def get_R_Delta(M_Delta,z,cosmology,Delta,crit):

    M_Delta = M_Delta*1e15
    const = constants()

    if crit == "critical":

        rho_c = cosmology.critical_density(z).value*1000.*const.mpc**3/const.solar
        R_Delta = (M_Delta/(Delta*4.*np.pi/3.*rho_c))**(1./3.)

    return R_Delta

class scalrel_websky:

    def __init__(self):

        self.cosmology = websky.cosmology_websky()
        self.A0 = 1
        self.B0 = 1
        self.M_pivot = 1
        self.const = constants()

    def get_y0(self,M_500,z):

        E = self.cosmology.H(z)/self.cosmology.H0
        y0 = 10.**self.A0*(M_500/self.M_pivot)**(1.+self.B0)*E**2

        return y0

    def get_theta_500(self,M_500,z,units="arcmin"):

        if units == "arcmin":

            theta_500 = get_theta_500_arcmin(M_500,z,self.cosmology)

        return theta_500

    def get_q(self,M_500,z,theta_vec,sigma_y_vec):

        sigma_y = np.interp(self.get_theta_500(M_500,z,units="arcmin"),theta_vec,sigma_y_vec)
        y0 = self.get_y0(M_500,z)

        return y0/sigma_y


class point_source:

    def __init__(self,pix,exp,beam="gaussian",freqs=None):

        point_source_map = np.zeros((pix.nx,pix.ny,len(freqs)))
        point_source_fft = np.zeros((pix.nx,pix.ny,len(freqs)))

        for i in range(0,len(freqs)):

            if beam =="real":

                if exp.name == "Planck":

                    ell_beam,fft_beam = exp.get_beam(i)

            ell = maps.rmap(pix).get_ell()
            beam_fft = np.interp(ell,ell_beam,fft_beam)
            beam_real = maps.get_ifft(beam_fft,pix).real

            point_source_map[:,:,i] = beam_real
            point_source_fft[:,:,i] = beam_fft

        self.point_source_map = point_source_map
        self.point_source_fft = point_source_fft

class gaussian_signal:

    def __init__(self,pix,A,sigma_arcmin,theta_misc=[0.,0.],n_freqs=1,sed=None):

        self.pix = pix
        self.A = A
        self.sigma_arcmin = sigma_arcmin
        self.sigma = self.sigma_arcmin/60./180.*np.pi
        self.theta_misc = theta_misc
        self.n_freqs = n_freqs
        self.sed = sed

    def get_signal_map(self):

        theta_map = maps.rmap(self.pix).get_distance_map_wrt_centre(self.theta_misc)
        signal_map = self.get_signal(theta_map)*self.A

        ret = np.zeros((self.pix.nx,self.pix.nx,self.n_freqs))

        for i in range(0,self.n_freqs):

            ret[:,:,i] = signal_map*self.sed[i]

        return ret

    def get_t_map_convolved(self,exp,beam="gaussian"):

        tmap = self.get_signal_map()
        n_freqs = exp.n_freqs

        tmap_convolved = maps.convolve_tmap_experiment(self.pix,tmap,exp,beam_type=beam)

        return tmap_convolved

    def get_signal(self,theta):

        return gaussian_1d(theta,0.,self.sigma)

    def get_norm(self):

        return self.get_signal(0.)*self.A



def gaussian_1d(x,mu,sigma):

    y = (x-mu)/sigma

    return np.exp(-0.5*y**2)/(np.sqrt(2.*np.pi*sigma**2))

class cib_model: #Modified blackbody, model used in Websky.

    def __init__(self,path="/Users/user/Desktop/",beta=None,T0=None):

        if beta is None:

            beta = 1.6

        if T0 is None:

            T0 = 20.7

        self.beta = beta
        self.gamma = 1.8
        self.alpha = 0.2
        self.T0 = T0
        self.nu0 = 10000.
        self.nu_pivot = 1#3e9
        self.path = path
        self.moments = {}

    #Input nu is experiment nu (i.e. at z = 0). Returns intensity in T_CMB units.

    def get_Theta(self,nu,z,units="T_CMB",beta=None):

        if beta is not None:

            self.beta = beta

        nup = nu*(1.+z)

        Theta_0 = (nup/self.nu_pivot)**self.beta*planckian(nup,self.T0*(1+z)**self.alpha)

        if units == "T_CMB":

            #Theta = Theta_0*MJysr_to_muK_factor(nu)/1e10
            exp = expt.planck_specs(path=self.path)
            Theta = Theta_0*exp.MJysr_to_muK_websky

        elif units == "SI":

            Theta = Theta_0

        return Theta

    def get_Theta_1d(self,nu,z,exp,units="T_CMB"):

        nup = nu*(1.+z)

        Theta_1_beta = self.get_Theta(nu,z,units=units)*np.log(nup/self.nu_pivot)

        const = constants()
        exponential = np.exp(const.h*nup/( const.k_B*self.T0*(1+z)**self.alpha))

        Theta_1_betaT = -2.*const.h*nup**(3.+self.beta)/(const.c_light**2*(exponential-1.)**2)*exponential*const.h*nup/const.k_B

        if units == "T_CMB":

            Theta_1_betaT = Theta_1_betaT*exp.MJysr_to_muK_websky

        self.Theta_1_betaT = Theta_1_betaT
        self.Theta_1_beta = Theta_1_beta
        self.moments["betaT"] = Theta_1_betaT
        self.moments["beta"] = Theta_1_beta

        return Theta_1_beta,Theta_1_betaT


    def get_sed_exp(self,exp,z,beta=None):

        nu_vec = exp.nu_eff

        return (nu_vec,self.get_Theta(nu_vec,z,beta=beta))

    def get_sed_1d_exp(self,exp,z,units="T_CMB"):

        nu_vec = exp.nu_eff

        Theta_1_beta,Theta_1_betaT = self.get_Theta_1d(nu_vec,z,exp,units=units)

        return Theta_1_beta,Theta_1_betaT

    def get_sed_exp_class_sz(self,exp):

        nu_vec = exp.nu_eff

#    [nu0,I0] = np.load("class_sz_cib_monopole_planck.npy",allow_pickle=True)
        [nu0,I0] = np.load("CIB_monopole_class_sz.npy",allow_pickle=True)
        I0 = I0*MJysr_to_muK_factor(nu0*1e9)
        I0 = np.interp(nu_vec,nu0*1e9,I0)/1e8

        return (nu_vec,I0)

    def get_sed_exp_empirical(self,exp,type="mean"):

        nu_vec = exp.nu_eff

        if type == "mean":

            I0 = np.load("sed_cib_planck_empirical.npy",allow_pickle=True)

        elif type == "cluster":

            I0 = np.load("sed_cib_planck_empirical_clusters.npy",allow_pickle=True)

        elif type == "cluster_only":

    #        sed_empirical_clusters,sed_empirical_random = np.load("sed_cib_planck_empirical_clusters_z03.npy",allow_pickle=True)
        #    sed_empirical_clusters,sed_empirical_random = np.load("sed_cib_planck_empirical_clusters_z03_masked.npy",allow_pickle=True)
            sed_empirical_clusters,sed_empirical_random = np.load("sed_cib_planck_empirical_clusters_z02_masked.npy",allow_pickle=True)
            I0 = sed_empirical_clusters-sed_empirical_random

        return (nu_vec,I0)

#nu in Hz, T in K

def planckian(nu,T):

    const = constants()
    planck = 2.*const.h*nu**3/(const.c_light**2*(np.exp(const.h*nu/(const.k_B*T))-1.))

    return planck

#nu in Hz, T_CMB in K. To go from Jy/sr to T_CMB, dvide by output of this function

def dBnudT(nu,T_CMB=constants().T_CMB):

    const = constants()

    x = const.h*nu/(const.k_B*T_CMB)
    f = x**2*np.exp(x)/(np.exp(x)-1.)**2.

    return 2.*const.k_B*nu**2/const.c_light**2*f

class ps_sz:

    def __init__(self):

        pars = {'h0':0.7, 'obh2':0.025,'och2':0.12,\
            'As':np.exp(3.06)*1e-10,'ns':0.9645,'mnu':0.06,\
            'flag_nu':True, 'flag_tll':False, 'mass_bias':1.5}
        tsz = pysz.tsz_cl()
        ell_arr = np.logspace(1,4.0,20)
        cl_yy, tll = tsz.get_tsz_cl(ell_arr,pars)
        cl = cl_yy[0]+cl_yy[1] #1 halo + 2 halo terms

        self.ell = np.arange(1,4000)
        self.cl = np.interp(np.log(ell),np.log(ell_arr),cl)

#From MJy/sr to muK_CMB. Input frequency in Hz, I in MJy/sr, returns in muK_CMB

def MJysr_to_muK(I,nu):

    const = constants()

    x = const.h*nu/(const.k_B*const.T_CMB)

    return 1.05e3*(np.exp(x)-1.)**2*np.exp(-x)*(nu/1e11)**(-4)*I

def MJysr_to_muK_factor(nu):

    const = constants()

    x = const.h*nu/(const.k_B*const.T_CMB)
    factor = 1.05e3*(np.exp(x)-1.)**2*np.exp(-x)*(nu/1e11)**(-4)

    return factor

class point_source:

    def __init__(self,experiment,beam_type="gaussian"):

        self.experiment = experiment
        self.beam_type = beam_type

    def get_t_map_convolved(self,pix):

        tem = np.zeros((pix.nx,pix.nx,self.experiment.n_freqs))

        for i in range(0,self.experiment.n_freqs):

            if self.beam_type == "gaussian":

                tem[:,:,i] = maps.eval_beam_real_map(pix,self.experiment.FWHM[i])

            elif self.beam_type == "real":

                tem[:,:,i] = maps.eval_real_beam_real_map(pix,self.experiment.get_beam(i))

            tem[:,:,i] = tem[:,:,i]/tem[pix.nx//2,pix.nx//2,i]

        return tem



import numpy as np
import pylab as pl
import healpy as hp
import expt
import pysm3
import pysm3.units as u
import healpy as hp
import numpy as np
import sphere
import cat
from matplotlib.patches import Circle
import astropy.cosmology as cp
import maps
import mmf
import os

class cosmology_websky:

    def __init__(self):

        self.Ob0 = 0.049
        self.Oc0 = 0.261
        self.Om0 = self.Ob0 + self.Oc0
        self.h      = 0.68
        self.ns     = 0.965
        self.sigma8 = 0.81
        self.cosmology = cp.FlatLambdaCDM(Om0=self.Om0,H0=self.h*100.,Ob0=self.Ob0)

class catalogue_websky_master:

    def __init__(self,mode="websky",path="/Users/user/Desktop/"):

        if mode == "websky":

            catalogue = np.load(path + "data/websky/halos_reduced.npy")
            catalogue_mass_converted = np.load(path + "data/websky/halos_converted_mass.npy")
            y0 = np.load(path + "data/websky/halos_y0.npy")

            self.x = catalogue[:,0]
            self.y = catalogue[:,1]
            self.z = catalogue[:,2]
            self.M200m = catalogue[:,3]
            self.redshift = catalogue[:,4]
            self.chi = catalogue[:,5]
            self.vrad = catalogue[:,6]
            self.r_s = catalogue[:,7]
            self.R200m = catalogue[:,8]/(200.)**(1./3.)
            self.theta_200m = catalogue[:,9]/(200.)**(1./3.)
            self.lon = catalogue[:,10]
            self.lat = catalogue[:,11]
            self.M500c = catalogue_mass_converted[:,0]
            self.R500c = catalogue_mass_converted[:,1]/(200.)**(1./3.)
            self.theta_500c = self.R500c/self.chi
            self.y0 = y0

        elif mode == "injected":

            catalogue = np.load(path + "data/websky/halos_injected_2.npy")
            self.x = catalogue[:,0]
            self.y = catalogue[:,1]
            self.z = catalogue[:,2]
            self.M200m = catalogue[:,3]
            self.redshift = catalogue[:,4]
            self.chi = catalogue[:,5]
            self.vrad = catalogue[:,6]
            self.r_s = catalogue[:,7]
            self.R200m = catalogue[:,8]
            self.theta_200m = catalogue[:,9]
            self.lon = catalogue[:,10]
            self.lat = catalogue[:,11]
            self.M500c = catalogue[:,12]
            self.R500c = catalogue[:,13]
            self.theta_500c = catalogue[:,14]
            self.y0 = catalogue[:,15]

    def select_threshold(self,select="mass",th=2.,m_min=1e14):

        if select == "mass":

            indices = np.where((self.M200m > m_min))

        elif select == "q":

            z_vec,M_500_vec = np.load("threshold_curve_" + str(th) + ".npy")
            th = self.M200m-np.interp(self.redshift,z_vec,M_500_vec*1e15)
            indices = np.where((th > 0.))

        return catalogue_websky(
        theta_200m=self.theta_200m[indices],lat=self.lat[indices],lon=self.lon[indices],M200m=self.M200m[indices],
        redshift=self.redshift[indices],R200m=self.R200m[indices],M500c=self.M500c[indices],R500c=self.R500c[indices],
        theta_500c=self.theta_500c[indices],y0=self.y0)

    def select_tile(self,i,pix,m_min=1e14,type="field",nside=8,select="mass",th=2.):

        theta_x,theta_y = sphere.get_xy(i,self.lon,self.lat,nside)

        if select == "mass":

            indices1 = np.where((self.M200m > m_min))

        elif select == "q":

            z_vec,M_500_vec = np.load("threshold_curve_" + str(th) + ".npy")
            th = self.M200m-np.interp(self.redshift,z_vec,M_500_vec*1e15)
            indices1 = np.where((th > 0.))

        theta_x = theta_x[indices1]
        theta_y = theta_y[indices1]
        theta_200m = self.theta_200m[indices1]
        M200m = self.M200m[indices1]
        lon = self.lon[indices1]
        lat = self.lat[indices1]
        redshift = self.redshift[indices1]
        R200m = self.R200m[indices1]
        M500c = self.M500c[indices1]
        R500c = self.R500c[indices1]
        theta_500c = self.theta_500c[indices1]
        y0 = self.y0[indices1]

        if type == "field":

            lx = pix.nx*pix.dx
            indices = np.where((theta_x < lx*0.5) & (theta_x > -lx*0.5) & (theta_y < lx*0.5) & (theta_y > -lx*0.5))

        elif type == "tile":

            pix_vec = hp.pixelfunc.ang2pix(nside,lon,lat,lonlat=True)
            indices = np.where((pix_vec == i))

        theta_x = theta_x + pix.nx*pix.dx*0.5
        theta_y = theta_y + pix.nx*pix.dx*0.5

        return catalogue_websky(theta_x=theta_x[indices],theta_y=theta_y[indices],
        theta_200m=theta_200m[indices],lat=lat[indices],lon=lon[indices],M200m=M200m[indices],
        redshift=redshift[indices],R200m=R200m[indices],M500c=M500c[indices],R500c=R500c[indices],
        theta_500c=theta_500c[indices],y0=y0[indices])

    def get_catalogue_measurement(self):

        negative = -np.ones(len(self.lat))*0.5

        return cat.cluster_catalogue(q_opt=negative
        ,y0=self.y0,theta_500=self.theta_500c*180.*60./np.pi,theta_x=self.theta_x
        ,theta_y=self.theta_y,
        pixel_ids=negative,lat=self.lat,
        lon=self.lon,m_500=self.M500c,z=self.redshift)

class catalogue_websky:

    def __init__(self,theta_x=np.empty(0),theta_y=np.empty(0),theta_200m=np.empty(0),
    lat=np.empty(0),lon=np.empty(0),theta_500c=np.empty(0),M200m=np.empty(0),M500c=np.empty(0),
    R200m=np.empty(0),R500c=np.empty(0),redshift=np.empty(0),y0=np.empty(0)):

        self.theta_x = theta_x
        self.theta_y = theta_y
        self.theta_200m = theta_200m
        self.theta_500c = theta_500c
        self.lat = lat
        self.lon = lon
        self.M200m = M200m
        self.M500c = M500c
        self.R200m = R200m
        self.R500c = R500c
        self.redshift = redshift
        self.theta_500 = theta_500c
        self.y0 = y0

    def get_catalogue_measurement(self):

        negative = -np.ones(len(self.lat))*0.5

        cat_return = cat.cluster_catalogue()
        cat_return.catalogue["q_opt"] = negative
        cat_return.catalogue["y0"] = self.y0
        cat_return.catalogue["theta_500"] = self.theta_500c*180.*60./np.pi
        cat_return.catalogue["theta_x"] = self.theta_x
        cat_return.catalogue["theta_y"] = self.theta_y
        cat_return.catalogue["pixel_ids"] = negative
        cat_return.catalogue["lat"] = self.lat
        cat_return.catalogue["lon"] = self.lon
        cat_return.catalogue["m_500"] = self.M500c
        cat_return.catalogue["z"] = self.redshift

        return cat_return

class websky_maps:

    def __init__(self):

        self.exp = expt.websky_specs()

    def get_tsz(self,i):

        return hp.read_map("/Users/user/Desktop/data/websky/tsz_" + str(self.exp.nu_eff_GHz[i])+ "_2048.fits",field=0)


#Produces a set of maps ALL IN muK

def preprocess_websky():

    exp = expt.websky_specs()

    #tSZ

    y_map = hp.read_map('/Users/user/Desktop/data/websky/tsz.fits',field=0)
    y_map = hp.pixelfunc.ud_grade(y_map,2048)

    #hp.mollview(y_map)
    #pl.show()

    for i in range(0,6):

        print(i)

        y_map_conv = hp.sphtfunc.smoothing(y_map,beam_window=expt.planck_specs().get_beam(i,ptf=False)[1])
        t_map = y_map_conv*exp.y2muK[i]

        #hp.mollview(t_map)
        #pl.show()

        hp.fitsfunc.write_map("/Users/user/Desktop/data/websky/tsz_" + str(exp.nu_eff_GHz[i])+ "_2048.fits",t_map)

    """
    #kSZ

    ks_map = hp.read_map('/Users/user/Desktop/data/websky/ksz.fits',field=0) + hp.read_map('/Users/user/Desktop/data/websky/ksz_patchy.fits',field=0)
    ks_map = hp.pixelfunc.ud_grade(ks_map,2048)

    hp.mollview(ks_map)
    pl.show()

    for i in range(0,6):

        print(i)

        t_map = hp.sphtfunc.smoothing(ks_map,beam_window=expt.planck_specs().get_beam(i,ptf=False)[1])

        #hp.mollview(t_map)
        #pl.show()

        hp.fitsfunc.write_map("/Users/user/Desktop/data/websky/ksz_" + str(exp.nu_eff_GHz[i])+ "_2048.fits",t_map)



    #CIB


    for i in range(0,6):

        print(i)

        cib_map = hp.read_map("/Users/user/Desktop/data/websky/cib_nu0" + str(exp.nu_eff_GHz[i]) + ".fits",field=0)

        std = np.std(cib_map)
        indices = np.where(cib_map > 3.*std)[0]
        cib_map[indices] = 0.

        cib_map = hp.pixelfunc.ud_grade(cib_map,2048)
        cib_map = cib_map*exp.MJysr2muK[i]
        t_map = hp.sphtfunc.smoothing(cib_map,beam_window=expt.planck_specs().get_beam(i,ptf=False)[1])

        hp.fitsfunc.write_map("/Users/user/Desktop/data/websky/cib_masked_" + str(exp.nu_eff_GHz[i])+ "_2048.fits",t_map)




    #CMB

    cmb_map = hp.read_map('/Users/user/Desktop/data/websky/cmb.fits',field=0)
    cmb_map = hp.pixelfunc.ud_grade(cmb_map,2048)

    #hp.mollview(ks_map)
    #pl.show()

    for i in range(0,6):

        print(i)

        t_map = hp.sphtfunc.smoothing(cmb_map,beam_window=expt.planck_specs().get_beam(i,ptf=False)[1])

        hp.fitsfunc.write_map("/Users/user/Desktop/data/websky/cmb_" + str(exp.nu_eff_GHz[i])+ "_2048.fits",t_map)

    """

    #tSZ injected

    """

    y_map = hp.read_map('/Users/user/Desktop/data/websky/tsz_injected_2048.fits',field=0)

    hp.mollview(y_map)
    pl.show()

    for i in range(0,6):

        print(i)

        y_map_conv = hp.sphtfunc.smoothing(y_map,beam_window=expt.planck_specs().get_beam(i,ptf=False)[1])
        t_map = y_map_conv*exp.y2muK[i]

        #hp.mollview(t_map)
        #pl.show()

        hp.fitsfunc.write_map("/Users/user/Desktop/data/websky/tsz_injected_" + str(exp.nu_eff_GHz[i])+ "_2048.fits",t_map)

    """

#All in muK

def get_galactic_foregrounds():

    exp = expt.websky_specs()

    dust = pysm3.Sky(nside=2048, preset_strings=["d1"])
    synchro = pysm3.Sky(nside=2048, preset_strings=["s1"])

    for i in range(0,6):

        dust_freq = dust.get_emission(exp.nu_eff_GHz[i]*u.GHz)
        dust_freq = dust_freq.to(u.uK_CMB,equivalencies=u.cmb_equivalencies(exp.nu_eff_GHz[i]*u.GHz))
        dust_freq_t = dust_freq[0]
        dust_freq_t = hp.sphtfunc.smoothing(dust_freq_t,beam_window=expt.planck_specs().get_beam(i,ptf=False)[1])

        hp.fitsfunc.write_map("/Users/user/Desktop/data/websky/dust_" + str(exp.nu_eff_GHz[i])+ "_2048.fits",dust_freq_t)
        #hp.mollview(dust_freq_t,max=1e2)
        #pl.show()

        synchro_freq = synchro.get_emission(exp.nu_eff_GHz[i]*u.GHz)
        synchro_freq = synchro_freq.to(u.uK_CMB,equivalencies=u.cmb_equivalencies(exp.nu_eff_GHz[i]*u.GHz))
        synchro_freq_t = synchro_freq[0]
        synchro_freq_t = hp.sphtfunc.smoothing(synchro_freq_t,beam_window=expt.planck_specs().get_beam(i,ptf=False)[1])
        hp.fitsfunc.write_map("/Users/user/Desktop/data/websky/synchro_" + str(exp.nu_eff_GHz[i])+ "_2048.fits",synchro_freq_t)

        #hp.mollview(synchro_freq_t,max=1e2)
        #pl.show()

#All in muK

def get_noise_maps(ptf_convolution=True):

    exp = expt.websky_specs()
    nside = 2048

    for i in range(0,6):

        print(i)
        if ptf_convolution == True:

            noise_map = sphere.get_white_noise_map_convolved_hp(nside,exp.noise_levels[i])
            hp.fitsfunc.write_map("/Users/user/Desktop/data/websky/noise_" + str(exp.nu_eff_GHz[i])+ "_2048.fits",noise_map)

        elif ptf_convolution == False:

            noise_map = sphere.get_white_noise_map_hp(nside,exp.noise_levels[i])
            hp.fitsfunc.write_map("/Users/user/Desktop/data/websky/noise_no_ptf_" + str(exp.nu_eff_GHz[i])+ "_2048.fits",noise_map)

class cutouts_websky:

    def __init__(self,path="/Users/user/Desktop/",inpaint_flag=False,mode="websky"):

        self.name = path + "maps/websky_maps/t_maps/"
        self.name_mask = path + "maps/planck_maps/planck_field_"
        self.mask_cib = False
        self.sigma_mask = 0.
        self.inpaint_flag = inpaint_flag
        self.nx = 1024
        self.l = 14.8  #in deg
        self.dx = self.l/self.nx/180.*np.pi
        self.pix = maps.pixel(self.nx,self.dx)
        self.nside_tile = 8
        self.n_tile = hp.nside2npix(self.nside_tile)
        self.mode = mode
        self.path = path

    def get_cutout_i(self,i):

        cutout_i = cutout_websky(i,name=self.name,name_mask=self.name_mask,
        mask_cib=self.mask_cib,sigma_mask=self.sigma_mask,inpaint_flag=self.inpaint_flag,mode=self.mode,path=self.path)

        return cutout_i

    def get_cutout_mean(self,i_min,i_max):

        dust = np.zeros((self.nx,self.nx,6))
        synchro = np.zeros((self.nx,self.nx,6))
        cib = np.zeros((self.nx,self.nx,6))
        ksz = np.zeros((self.nx,self.nx,6))
        noise = np.zeros((self.nx,self.nx,6))
        cmb = np.zeros((self.nx,self.nx,6))
        cib_masked = np.zeros((self.nx,self.nx,6))
        noise_no_ptf = np.zeros((self.nx,self.nx,6))

        for i in range(i_min,i_max):

            cutout_i = cutout_websky(i,name=self.name,name_mask=self.name_mask,
            mask_cib=self.mask_cib,sigma_mask=self.sigma_mask,inpaint_flag=self.inpaint_flag,mode=self.mode)

            dust = dust + cutout_i.dust
            synchro = synchro + cutout_i.synchro
            cib = cib + cutout_i.cib
            ksz = ksz + cutout_i.ksz
            noise = noise + cutout_i.noise
            cmb = cmb + cutout_i.cmb
            cib_masked = cib_masked + cutout_i.cib_masked
            #noise_no_ptf = noise_no_ptf + cutout_i.noise_no_ptf

        cutout_i.dust = dust/float(i_max-i_min)
        cutout_i.synchro = synchro/float(i_max-i_min)
        cutout_i.cib = cib/float(i_max-i_min)
        cutout_i.ksz = ksz/float(i_max-i_min)
        cutout_i.noise = noise/float(i_max-i_min)
        cutout_i.cmb = cmb/float(i_max-i_min)
        cutout_i.cib_masked = cib_masked/float(i_max-i_min)
    #    cutout_i.noise_no_ptf = noise_no_ptf/float(i_max-i_min)

        return cutout_i

#everything in muK

class cutout_websky:

    def __init__(self,i,name="maps/websky_maps/t_maps/",
    name_mask="maps/planck_maps/planck_field_",mask_cib=False,sigma_mask=None,inpaint_flag=False,mode="real",
    path = "/Users/user/Desktop/"):

        #name_mask = path + name_mask
        #name = path + name

        [mask_galaxy,mask_point,mask_tile] = np.load(name_mask + str(i) + "_mask.npy")

        self.mask_galaxy = mask_galaxy
        self.mask_point = mask_point
        self.mask_tile = mask_tile

        self.dust = np.load(name + "_dust_" + str(i) + "_tmap.npy")[0,:,:,:]
        self.synchro = np.load(name + "_synchro_" + str(i) + "_tmap.npy")[0,:,:,:]
        self.cib = np.load(name + "_cib_" + str(i) + "_tmap.npy")[0,:,:,:]

        if mode == "websky":

            self.tsz = np.load(name + "_tsz_" + str(i) + "_tmap.npy")[0,:,:,:]

        elif mode == "injected": #can be injected or injected_flat

            self.tsz = np.load(name + "_tsz_" + mode + "_" + str(i) + "_tmap.npy")[0,:,:,:]

        elif mode == "injected_flat":

            self.tsz = np.load(name + "_tsz_" + mode + "_" + str(i) + "_tmap.npy")

        self.ksz = np.load(name + "_ksz_" + str(i) + "_tmap.npy")[0,:,:,:]
        self.noise = np.load(name + "_noise_" + str(i) + "_tmap.npy")[0,:,:,:]
        self.cmb = np.load(name + "_cmb_" + str(i) + "_tmap.npy")[0,:,:,:]
        self.cib_masked = np.load(name + "_cib_masked_" + str(i) + "_tmap.npy")[0,:,:,:]
    #    self.noise_no_ptf = np.load(name + "_noise_no_ptf_" + str(i) + "_tmap.npy")[0,:,:,:]
        self.mask_point_cib = np.load(path + "maps/websky_maps/cib_mask_" + str(i) + ".npy")

        if mask_cib == True:

            self.cib = maps.mask_sigma_freq(self.cib,sigma_mask,inpaint_flag=inpaint_flag)

        self.nx = 1024
        self.l = 14.8  #in deg
        self.dx = self.l/self.nx/180.*np.pi
        self.pix = maps.pixel(self.nx,self.dx)

    def get_tmap(self,mode="all"):

        if mode == "all":

            ret = self.ksz + self.dust + self.synchro + self.cmb + self.noise + self.tsz # + self.cib_masked #

        elif mode == "extrag":

            ret = self.tsz + self.ksz + self.cmb + self.noise #+ self.cib_masked

        elif mode == "extrag_noise":

            ret = self.ksz + self.cmb + self.noise

        elif mode == "cmb+noise":

            ret = self.tsz + self.cmb + self.noise

        elif mode == "cmb+noise-tsz":

            ret = self.cmb + self.noise

        elif mode == "cmb+noise+ksz":

            ret = self.tsz + self.cmb + self.noise

        elif mode == "all-tsz":

            ret = self.ksz + self.dust + self.synchro + self.cmb + self.noise + self.cib_masked

        return ret



class detections_websky:

    def __init__(self,i_min=0,i_max=100,mode="all",indices=None):

        self.catalogue_obs = cat.cluster_catalogue()
        self.catalogue_obs_noit = cat.cluster_catalogue()
        self.catalogue_true = cat.cluster_catalogue()

        cutout = cutout_websky(0)
        pix = cutout.pix

        cutouts = cutouts_websky()

        apply_mask = True

        if indices is None:

            indices = range(i_min,i_max)
            self.sigma_matrix = np.zeros((i_max-i_min,5))

        else:

            self.sigma_matrix = np.zeros((len(indices),5))

        j = 0

        for i in indices:

            if mode == "all":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_all_" + str(i) + ".npy"

            if mode == "all+cib":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_cib_" + str(i) + ".npy"

            if mode == "all+cib_mean":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_cib_mean_" + str(i) + ".npy"

            if mode == "all+cib_random_2200":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_cib_random_2200_" + str(i) + ".npy"

            if mode == "all+cib_random":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_cib_random_" + str(i) + ".npy"

            elif mode == "cmb+noise":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_cmb_" + str(i) + ".npy"

            if mode == "cmb+noise_tnoi":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_cmb_tnoi_" + str(i) + ".npy"

            if mode == "all+cib_random_tnoi":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_cib_random_tnoi_" + str(i) + ".npy"

            if mode == "cmb+noise_tnoi_nops":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_cmb_tnoi_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "all+cib_random_tnoi_nops":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_all_cmb_random_ßtnoi_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "cmb+noise_tnoi_nops_l2":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_cmb_tnoi_nops_l2_" + str(i) + ".npy"
                apply_mask = False

            if mode == "cmb+noise_tnoi_nops_newf":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_cmb_tnoi_nops_newf_" + str(i) + ".npy"
                apply_mask = False

            if mode == "all+cib_random_tnoi_nops_newf":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_all_cib_random_tnoi_nops_newf_" + str(i) + ".npy"
                apply_mask = False

            if mode == "cmb+noise_tnoi_nops_newf_fixed":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_cmb_tnoi_nops_newf_fixed_" + str(i) + ".npy"
                apply_mask = False

            if mode == "noise_tnoi_nops_newf_fixed":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_wn_tnoi_nops_newf_fixed_" + str(i) + ".npy"
                apply_mask = False

            if mode == "all_cib_random_tnoi_nops_newf_fixed":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_all_cib_random_tnoi_nops_newf_fixed_" + str(i) + ".npy"
                apply_mask = False

            if mode == "wn_tnoi_nops_newf":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_wn_tnoi_nops_newf_fixed_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_wn_tnoi_nops_newf":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_wn_tnoi_nops_newf_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_all+cib_random_tnoi_nops_newf": #0-100 and 667-768

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_all_cib_random_tnoi_newf_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_cmb+noise_tnoi_nops_newf":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_cmb_tnoi_newf_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_wn_tnoi_nops_newf":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_wn_tnoi_newf_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_all_tnoi_nops_newf":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_all_tnoi_newf_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "all_tnoi_nops_newf":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_all_tnoi_nops_newf_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_all+cib_random_mean_tnoi_nops_newf":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_all+cib_random_mean_tnoi_newf_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_wn_tnoi_nops_newf_nof":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_wn_tnoi_newf_nops_nof_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_wn_tnoi_nops_newf_fixed":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_wn_tnoi_newf_nops_fixed_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_wn_tnoi_nops_newf_noptf":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_wn_tnoi_newf_nops_noptf_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_cmb+noise_tnoi_nops_newf_fixed":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_cmb_tnoi_newf_nops_fixed_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_all+cib_random_tnoi_nops_newf_fixed":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_all+cib_random_tnoi_newf_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_all+cib_random_tnoi_nops_newf_fixed_4ch":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_all+cib_random_tnoi_newf_nops_4ch_" + str(i) + ".npy"
                apply_mask = False

            if mode == "all_cib_random_tnoi_newf":

                name = "/Users/user/Desktop/catalogues/websky/catalogue_all+cib_random_tnoi_newf_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_all_cib_random_cmb_comp":

                name = "/Users/user/Desktop/catalogues/websky/component_cmb_inj_all+cib_random_tnoi_newf_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_all_cib_random_noi_comp":

                name = "/Users/user/Desktop/catalogues/websky/component_noi_inj_all+cib_random_tnoi_newf_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_all_cib_random_cib_comp":

                name = "/Users/user/Desktop/catalogues/websky/component_cib_inj_all+cib_random_tnoi_newf_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_all_cib_random_synchro_comp":

                name = "/Users/user/Desktop/catalogues/websky/component_synchro_inj_all+cib_random_tnoi_newf_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_all_cib_random_dust_comp":

                name = "/Users/user/Desktop/catalogues/websky/component_dust_inj_all+cib_random_tnoi_newf_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_all_cib_random_ksz_comp":

                name = "/Users/user/Desktop/catalogues/websky/component_ksz_inj_all+cib_random_tnoi_newf_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_all+cib_random_nops_newf_fixed": #0-100 and 667-768

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_all+cib_random_nops_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_all+cib_random_nops_newf": #0-100 and 667-768

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_all+cib_random_nops_find_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_extr+cib_random_tnoi_nops_newf": #0-768

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr+cib_random_tnoi_nops_find_"  + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_extr+cib_random_tnoi_nops_newf_qttrue": #0-768

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr+cib_random_tnoi_nops_find_qttrue_"  + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_extr+cib_random_tnoi_nops_newf_qttrue_flat": #0-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr+cib_random_tnoi_nops_find_qttrue_flat_"  + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_extr+cib_random_tnoi_nops_newf_flat": #0-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr+cib_random_tnoi_nops_find_flat_"  + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_extr+cib_random_tnoi_nops_newf_tsubgrid_flat": #0-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr+cib_random_tnoi_nops_find_tsubgrid_flat_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_extr+cib_random_tnoi_nops_newf_tsubgrid_findnosubgrid_flat": #0-100, 667-768

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr+cib_random_tnoi_nops_find_tsubgrid_findnosubgrid_flat_" + str(i) + ".npy"
                apply_mask = False

            if mode == "extrag_cib_tnoi_nops_newf_fixed": #0-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_extrag_cib_tnoi_nops_newf_fixed_" + str(i) + ".npy"
                apply_mask = False

            if mode == "extrag_cib_random_tnoi_nops_newf_fixed": #0-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_extrag_cib_random_tnoi_nops_newf_fixed_" + str(i) + ".npy"
                apply_mask = False

            if mode == "extrag_cib_random_tnoi_nops_newf_find_tsubgrid_findnosubgrid": #0-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_extrag_cib_tnoi_nops_newf_find_tsubgrid_findnosubgrid_" + str(i) + ".npy"
                apply_mask = False

            if mode == "extrag_cib_random_tnoi_nops_newf_find_tsubgrid_findnosubgrid_peak": #0-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_extrag_cib_tnoi_nops_newf_find_tsubgrid_findnosubgrid_peak_" + str(i) + ".npy"
                apply_mask = False

            if mode == "extrag_cib_random_nops_newf_find_tsubgrid_findnosubgrid_peak": #0-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_extrag_cib_nops_newf_find_tsubgrid_findnosubgrid_" + str(i) + ".npy"
                apply_mask = False

            if mode == "extrag_cib_random_nops_newf_tsubgrid_fixed": #0-100, 667-678

                name = "/Users/user/Desktop/catalogues/websky/catalogue_extrag_cib_nops_newf_find_tsubgrid_fixed_" + str(i) + ".npy"
                apply_mask = False

            if mode == "extrag_cib_random_nops_newf_tsubgrid_fixed_tszps_subtracted": #0-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_extrag_cib_nops_newf_find_tsubgrid_fixed_tszps_subtracted_" + str(i) + ".npy"

            if mode == "extrag_cib_random_nops_newf_tsubgrid_fixed_tszps_subtracted_noit": #0-200

                name = "/Users/user/Desktop/catalogues/websky/catalogue_extrag_cib_nops_newf_find_tsubgrid_fixed_tszps_subtracted_noit_" + str(i) + ".npy"
                apply_mask = False


            if mode == "extrag_cib_random_nops_newf_tsubgrid_fixed_5": #0-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_extrag_cib_nops_newf_find_tsubgrid_fixed_5_" + str(i) + ".npy"
                apply_mask = False

            if mode == "extrag_cib_random_nops_newf_tsubgrid_fixed_noit": #0-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_extrag_cib_nops_newf_find_tsubgrid_fixed_noit_" + str(i) + ".npy"

                if i >= 100:

                    name = "/Users/user/Desktop/catalogues/websky/extrag_cib_random_nops_newf_tsubgrid_fixed_noit_" + str(i) + ".npy"

                apply_mask = False

            if mode == "extrag_cib_random_nops_newf_tsubgrid_fixed_tszps_subtracted_masked": #0-200 100-200 at q=5

                name = "/Users/user/Desktop/catalogues/websky/extrag_cib_random_nops_newf_tsubgrid_fixed_tszps_subtracted_masked_" + str(i) + ".npy"
                apply_mask = False

            if mode == "extrag_cib_random_nops_newf_tsubgrid_fixed_tszps_subtracted_masked_b":

                name = "/Users/user/Desktop/catalogues/websky/extrag_cib_random_nops_newf_tsubgrid_fixed_tszps_subtracted_masked_b_" + str(i) + ".npy"
                apply_mask = False

            if mode == "extrag_cib_random_nops_newf_tsubgrid_fixed_5_subtracted_masked": #0-200

                name = "/Users/user/Desktop/catalogues/websky/catalogue_extrag_cib_nops_newf_find_tsubgrid_fixed_5_subtract_masked_" + str(i) + ".npy"
                apply_mask = False

            if mode == "true_tsz_random_tsubgrid_fixed": #100-200

                name = "/Users/user/Desktop/catalogues/websky/true_tsz_random_tsubgrid_fixed" + str(i) + ".npy"
                apply_mask = False

            if mode == "true_tsz_random_tsubgrid_fixed_masked": #100-200

                name = "/Users/user/Desktop/catalogues/websky/true_tsz_random_tsubgrid_fixed_masked_" + str(i) + ".npy"
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_fixed_subgrid": #00-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr_cib_random_nops_fixed_nosubgrid_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_fixed_nosubgrid": #00-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr_cib_random_nops_fixed_subgridnot_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_fixed_qtrue": #00-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr_cib_random_nops_fixed_qtrue_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_fixed_subgrid_qtrue": #00-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr_cib_random_nops_fixed_subgrid_qtrue_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_find_subgrid_qtrue": #00-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr_cib_random_nops_find_subgrid_qtrue_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_fixed_subgrid_noit": #00-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr_cib_random_nops_fixed_noit_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            #From here on: new apodisation

            if mode == "inj_extr_cib_random_nops_tnoi_find_apodnew": #00-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr_cib_random_nops_tnoi_find_apodnew_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_tnoi_fixed_apodnew": #00-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr_cib_random_nops_tnoi_fixed_apodnew_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_it_find_apodnew": #00-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr_cib_random_nops_it_find_apodnew_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_noit_find_apodnew": #00-100

                name = "/Users/user/Desktop/catalogues/websky/catalogue_inj_extr_cib_random_nops_noit_find_apodnew_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_tnoi_find_apodnew_dbscan": #00-100

                name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_tnoi_find_apodnew_dbscan_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_it_fixed_apodnew": #00-100

                name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_it_fixed_apodnew_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_noit_fixed_apodnew": #00-100

                name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_noit_fixed_apodnew_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_tnoi_find_apodnew": #00-100

                name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_tnoi_find_apodnew_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_tnoi_fixed_qtrue_apodnew": #00-100

                name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_tnoi_fixed_qtrue_apodnew_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            #End of new apodisation

            if mode == "inj_extr_cib_random_nops_tnoi_find_apodold_dbscan": #00-100

                name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_tnoi_find_apodold_dbscan_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_tnoi_find_apodold": #00-100

                name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_tnoi_find_apodold_dbscan_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_tnoi_find_apodold_15": #all

                name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_tnoi_find_apodold_15_" + str(i) + ".npy" #it's fixed
                apply_mask = False

                if i >= 100:

                    name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_tnoi_find_apodold_" + str(i) + ".npy" #it's fixed
                    apply_mask = False

            if mode == "inj_extr_cib_random_nops_noit_fixed_apodold": #all

                name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_noit_fixed_apodold_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_it_find_apodold_15": #all


                name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_it_find_apodold_" + str(i) + ".npy" #it's fixed


                apply_mask = False

            if mode == "inj_extr_cib_random_nops_it_fixed_apodold": #all

                name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_it_fixed_apodold_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_tnoi_fixed_apodold": #all

                name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_tnoi_fixed_apodold_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_noit_find_apodold_15": #all

                name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_noit_find_apodold_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if mode == "inj_extr_cib_random_nops_it_fixed_apodold_truemasked": #all

                name = "/Users/user/Desktop/catalogues/websky/inj_extr_cib_random_nops_it_fixed_apodold_truemasked_" + str(i) + ".npy" #it's fixed
                apply_mask = False

            if os.path.isfile(name) == True:

                #print(i)

                results_new,a = np.load(name,allow_pickle=True)
                results_new.get_lonlat(i,pix)
                results_new.results_true.pixel_ids = np.ones(len(results_new.results_true.q_opt))*i #comment out when this is sorted out in code
                results_new.results_true.get_lonlat(8,i,pix) #idem
                self.catalogue_obs.append(results_new.results_refined)
                self.catalogue_obs_noit.append(results_new.results_refined_noit)

                results_true = results_new.results_true

                if apply_mask == True:

                    mask_select = np.load("/Users/user/Desktop/maps/websky_maps/mask_select/mask_select_" + str(i) + ".npy")
                    results_true = cat.apply_mask_select(results_true,mask_select,pix)

                self.catalogue_true.append(results_true)

                #print(i,len(results_new.results_true.q_opt),len(results_new.results_refined.q_opt))
                #if len(results_new.results_true.q_opt) < 20:

                #    print(i,len(results_new.results_true.q_opt))

                #self.sigma_matrix[j,:] = results_new.sigma_vec
                self.theta_500_sigma = results_new.theta_500_sigma

                j += 1

class detections_websky_paper:

    def __init__(self,data_label,noise_find_label,suffix="",i_min=0,i_max=100,
    indices=None,cib="random",name=None,path="/Users/user/Desktop/catalogues_def/websky/"):

        if cib == "random":

            self.prename = "paper_extr_cib_random_nops"

        elif cib == "":

            self.prename = "paper_extr_cib_nops"

        self.data_label = data_label
        self.noise_find_label = noise_find_label

        if name is None:

            full_name = path + self.prename + "_" + self.noise_find_label + suffix + "_"  + self.data_label

        else:

            full_name = path + name

        print(full_name)

        self.catalogue_obs = cat.cluster_catalogue()
        self.catalogue_obs_noit = cat.cluster_catalogue()
        self.catalogue_true = cat.cluster_catalogue()

        cutout = cutout_websky(0)
        pix = cutout.pix

        if indices is None:

            indices = range(i_min,i_max)
            self.sigma_matrix = np.zeros((i_max-i_min,15))

        else:

            self.sigma_matrix = np.zeros((len(indices),15))

        j = 0

        for i in indices:

            name = full_name + "_"  + str(i) + ".npy"

            if os.path.isfile(name) == True:

                #print(i)

                results_new,a = np.load(name,allow_pickle=True)
                cat.convert_results(results_new)
                results_new.get_lonlat(i,pix)
                results_new.results_true.catalogue["pixel_ids"] = np.ones(len(results_new.results_true.catalogue["q_opt"]))*i #comment out when this is sorted out in code
                results_new.results_true.get_lonlat(8,i,pix) #idem

                self.catalogue_obs.append(results_new.results_refined,append_keys="new")
                self.catalogue_obs_noit.append(results_new.results_refined_noit,append_keys="new")

                results_true = results_new.results_true

                self.catalogue_true.append(results_true,append_keys="new")

            #    if len(results_new.results_refined.catalogue["q_opt"]) != len(results_new.results_true.catalogue["q_opt"]):

                #    print(i,len(results_new.results_refined.catalogue["q_opt"]),len(results_new.results_true.catalogue["q_opt"]))
                #if len(results_new.results_true.q_opt) < 20:

                #    print(i,len(results_new.results_true.q_opt))

                self.sigma_matrix[j,:] = results_new.sigma_vec
                self.theta_500_sigma = results_new.theta_500_sigma

                j += 1

class detections_planck:

    def __init__(self,name,i_min=0,i_max=100,indices=None):

        self.name = name

        self.catalogue_obs = cat.cluster_catalogue()
        self.catalogue_obs_noit = cat.cluster_catalogue()

        cutout = cutout_websky(0)
        pix = cutout.pix

        if indices is None:

            indices = range(i_min,i_max)
            self.sigma_matrix = np.zeros((i_max-i_min,15))

        else:

            self.sigma_matrix = np.zeros((len(indices),15))

        j = 0

        full_name = "/Users/user/Desktop/catalogues_def/websky/" + name

        for i in indices:

            name = full_name + "_"  + str(i) + ".npy"

            if os.path.isfile(name) == True:

                #print(i)

                results_new,a = np.load(name,allow_pickle=True)
                results_new.get_lonlat(i,pix)
                self.catalogue_obs.append(results_new.results_refined)
                self.catalogue_obs_noit.append(results_new.results_refined_noit)

                self.sigma_matrix[j,:] = results_new.sigma_vec
                self.theta_500_sigma = results_new.theta_500_sigma

                j += 1

class detections_websky_processed:

    def __init__(self,data_label,noise_find_label,name_addition="",pixel_ids=None,cib="random",name=None):

            if cib == "random":

                prename = "paper_extr_cib_random_nops_" + noise_find_label + "_" + name_addition + data_label

            elif cib == "":

                prename = "paper_extr_cib_nops_" + noise_find_label + "_" + name_addition + data_label

            if name is not None:

                prename = name

            full_name = "/Users/user/Desktop/catalogues_def/processed_websky_paper/" + prename  + ".npy"
            (catalogue_obs,catalogue_true,metadata) = np.load(full_name,allow_pickle=True)

            print(full_name)

            catalogue_obs = cat.convert_catalogue(catalogue_obs)
            catalogue_true = cat.convert_catalogue(catalogue_true)

            self.catalogue_obs = catalogue_obs
            self.catalogue_true = catalogue_true

            if pixel_ids is not None:

                catalogue_obs_ret = cat.cluster_catalogue()
                catalogue_true_ret = cat.cluster_catalogue()

                for i in range(0,len(pixel_ids)):

                    indices = np.where(self.catalogue_obs.catalogue["pixel_ids"] == pixel_ids[i])
                    catalogue_obs_ret.append(cat.get_catalogue_indices(self.catalogue_obs,indices))
                    catalogue_true_ret.append(cat.get_catalogue_indices(self.catalogue_true,indices))

                self.catalogue_obs = catalogue_obs_ret
                self.catalogue_true = catalogue_true_ret

class detections_planck_processed:

    def __init__(self,name,pixel_ids=None):

        full_name = "/Users/user/Desktop/catalogues_def/processed_planck/" + name  + ".npy"
        (catalogue_obs,catalogue_obs_noit,metadata) = np.load(full_name,allow_pickle=True)

        self.catalogue_obs = catalogue_obs
        self.catalogue_obs_noit = catalogue_obs_noit

        if pixel_ids is not None:

            catalogue_obs_ret = cat.cluster_catalogue()
            catalogue_obs_noit_ret = cat.cluster_catalogue()

            for i in range(0,len(pixel_ids)):

                indices = np.where(self.catalogue_obs.catalogue["pixel_ids"] == pixel_ids[i])
                catalogue_obs_ret.append(cat.get_catalogue_indices(self.catalogue_obs,indices))
                catalogue_obs_noit_ret.append(cat.get_catalogue_indices(self.catalogue_obs_noit,indices))

            self.catalogue_obs = catalogue_obs_ret
            self.catalogue_obs_noit = catalogue_obs_noit_ret


def draw_circles(catalogue_tile,img,pix,scaling=1.,theta_500_units="rad",
save_name=None,plot=True,cmap=pl.get_cmap("RdYlBu"),pixel_id=0,title=None):

    nx = pix.nx
    dx = pix.dx
    extent_arcmin = nx*dx*180.*60./np.pi

    theta_x = catalogue_tile.catalogue["theta_x"]
    theta_y = catalogue_tile.catalogue["theta_y"]
    theta_500 = catalogue_tile.catalogue["theta_500"]#c
    #theta_500 = catalogue_tile.theta_200m

    if theta_500_units == "arcmin":

        theta_500 = theta_500/180./60.*np.pi

    theta_500 = theta_500*180.*60./np.pi

    pl.rc('text', usetex=True)
    pl.rc('font', family='serif')
    fig,ax = pl.subplots(1)
    ax.set_aspect('equal')
    ax.set_xlabel("arcmin")
    ax.set_ylabel("arcmin")

    if title is not None:

        ax.set_title(title)

    img_file = ax.imshow(img,cmap=cmap,extent=[0,extent_arcmin,0,extent_arcmin],interpolation=None)
    #fig.colorbar(img_file,ax=ax)

    xc = theta_x*180.*60./np.pi
    yc = ((theta_y))*180.*60./np.pi
    r = theta_500/scaling#/dx/scaling

    for i in range(0,len(theta_500)):

        circ = Circle((xc[i],yc[i]),r[i],fill=False)
        ax.add_patch(circ)

    if save_name is not None:

        pl.savefig(save_name)

    if plot == True:

        pl.show()

    pl.close()


def draw_circles_subplot(catalogue_tile,ax,pix,scaling=1.,theta_500_units="rad"):

    nx = pix.nx
    dx = pix.dx

    theta_x = catalogue_tile.catalogue["theta_x"]
    theta_y = catalogue_tile.catalogue["theta_y"]
    theta_500 = catalogue_tile.catalogue["theta_500"]#c
    #theta_500 = catalogue_tile.theta_200m

    if theta_500_units == "arcmin":

        theta_500 = theta_500/180./60.*np.pi

    ax.set_aspect('equal')

    xc = theta_x/dx
    yc = nx - (theta_y/dx)
    r = theta_500/dx/scaling

    for i in range(0,len(theta_500)):

        circ = Circle((xc[i],yc[i]),r[i],fill=False)
        ax.add_patch(circ)

def draw_circles_lonlat(catalogue_tile,img,pix,scaling=1.,theta_500_units="rad",
save_name=None,plot=True,cmap=pl.get_cmap("RdYlBu"),pixel_id=0):

    nx = pix.nx
    dx = pix.dx

    theta_x = catalogue_tile.catalogue["theta_x"]
    theta_y = catalogue_tile.catalogue["theta_y"]
    theta_500 = catalogue_tile.catalogue["theta_500"]#c
    #theta_500 = catalogue_tile.theta_200m


    if theta_500_units == "arcmin":

        theta_500 = theta_500/180./60.*np.pi

    fig,ax = pl.subplots(1)
    ax.set_aspect('equal')

    (lon_bl,lat_bl,lon_br,lat_br,lon_tl,lat_tl,lon_tr,lat_tr) = sphere.get_field_limits(8,pixel_id,nx,dx)

    ax.imshow(img,cmap=cmap,extent=[lon_bl,lon_br,lat_bl,lat_tl],interpolation='none')#,vmin=-3, vmax=8)
    ax.set_aspect(abs((lon_bl-lon_br)/(lat_bl-lat_tl)))
    xc = theta_x/dx
    yc = nx - (theta_y/dx)
    r = theta_500/dx/scaling

    xc = catalogue_tile.catalogue["lon"]
    yc = catalogue_tile.catalogue["lat"]
    r = theta_500/np.pi*180.

    print(xc)

    print(r)


    for i in range(0,len(theta_500)):

        circ = Circle((xc[i],yc[i]),r[i],fill=False)
        ax.add_patch(circ)

    if save_name is not None:

        pl.savefig(save_name)

    if plot == True:

        pl.show()

    pl.close()
